
import numpy as np
from McUtils.Scaffolding import ParameterManager

from .ColbertMiller import PolarDVR, RingDVR, CartesianDVR
from .DirectProduct import DirectProductDVR
from .Extensions import SelfConsistentDVR, PotentialOptimizedDVR

__all__ = [
    "DVR"
]

class DVRConstructor:

    _domain_map = None
    @classmethod
    def load_domain_map(cls):

        return {
            (0, np.pi): PolarDVR,
            (0, 2*np.pi): RingDVR,
            None: CartesianDVR
        }
    @classmethod
    def infer_DVR_type(cls, domain):
        if cls._domain_map is None:
            cls._domain_map = cls.load_domain_map()
        for k,v in cls._domain_map.items():
            if k is not None:
                if np.allclose(k, domain):
                    return v
        else:
            return cls._domain_map[None]

    @classmethod
    def construct(cls,
                  domain=None,
                  divs=None,
                  potential_function=None,
                  g=None,
                  g_deriv=None,
                  mass=None,
                  po_divs=25,
                  classes=None,
                  scf=False,
                  potential_optimize=False,
                  logger=True,
                  **base_opts
                  ):

        # dispatches based on domain to construct the appropriate DVR
        if domain is None or divs is None:
            raise ValueError("can't have `None` for `domain` or `divs`")
        if isinstance(domain[0], (int, float, np.integer, np.floating)): # 1D
            domain = [domain]
            divs = [divs]
            if mass is not None:
                mass = [mass]
        if classes is None:
            classes = [None] * len(domain)
        if g is not None:
            if callable(g):
                subg = [g]
            else:
                subg = [g[i][i] for i in range(len(g))]
            mass = [None] * len(subg)
            if g_deriv is None:
                g_deriv = [None] * len(subg)
        else:
            subg = [None]*len(mass)
            g_deriv = [None]*len(mass)

        if isinstance(po_divs, int):
            po_divs = [po_divs]*len(mass)

        ndim = len(list(zip(domain, divs, classes, mass, subg, g_deriv)))
        if ndim == 1:
            classes = [
                cls.infer_DVR_type(r) if c is None else c
                for r, n, c, m, sg, gd, nwf in zip(domain, divs, classes, mass, subg, g_deriv, po_divs)
            ]

            dvr = classes[0](
                domain=domain[0],
                divs=divs[0],
                potential_function=potential_function,
                g=subg[0],
                mass=mass[0],
                g_deriv=g_deriv[0],
                logger=logger,
                **base_opts
            )
        else:
            dvrs_1D = [
                cls.infer_DVR_type(r)(domain=r, divs=n, mass=m, g=sg, g_deriv=gd, num_wfns=nwf) if c is None else c(domain=r, divs=n)
                for r, n, c, m, sg, gd, nwf in zip(domain, divs, classes, mass, subg, g_deriv, po_divs)
            ]
            dvr = DirectProductDVR(
                dvrs_1D,
                domain=domain,
                divs=divs,
                potential_function=potential_function,
                g=g,
                mass=mass,
                g_deriv=g_deriv,
                logger=logger if not potential_optimize or scf else None,
                **ParameterManager(base_opts).exclude((SelfConsistentDVR, PotentialOptimizedDVR))
            )
            
            if potential_optimize or scf:
                if potential_optimize and scf is False:
                    dvr = PotentialOptimizedDVR.from_minimum(dvr,
                                                         logger=logger,
                                                         **ParameterManager(base_opts).filter(PotentialOptimizedDVR)
                                                         )
                else:
                    dvr = SelfConsistentDVR(dvr, logger=logger if not potential_optimize else None, **ParameterManager(base_opts).filter(SelfConsistentDVR))
                    if potential_optimize:
                        dvr = PotentialOptimizedDVR.from_scf(dvr,
                                                             logger=logger,
                                                             **ParameterManager(base_opts).filter(PotentialOptimizedDVR)
                                                             )
        return dvr


def DVR(
        domain=None,
        divs=None,
        classes=None,
        potential_function=None,
        g=None,
        g_deriv=None,
        scf=False,
        potential_optimize=False,
        **base_opts
):
    """
    Constructs a DVR object

    :param domain:
    :type domain:
    :param divs:
    :type divs:
    :param classes:
    :type classes:
    :param potential_function:
    :type potential_function:
    :param g:
    :type g:
    :param g_deriv:
    :type g_deriv:
    :param base_opts:
    :type base_opts:
    :return:
    :rtype:
    """

    return DVRConstructor.construct(
        domain=domain,
        divs=divs,
        classes=classes,
        potential_function=potential_function,
        g=g,
        g_deriv=g_deriv,
        scf=scf,
        potential_optimize=potential_optimize,
        **base_opts
    )


"""
Redoes what was originally PyDVR but in the _right_ way using proper subclassing and abstract properties
"""

import abc, numpy as np, scipy.sparse as sp, scipy.interpolate as interp

from McUtils.Data import UnitsData
from McUtils.Scaffolding import Logger, NullLogger

__all__ = ["BaseDVR", "DVRResults", "DVRException"]

class BaseDVR(metaclass=abc.ABCMeta):
    """
    Provides the abstract interface for creating a
    convenient runnable DVR that can be cleanly subclassed to provide
    extensions
    """

    def __init__(self,
                 domain=None,
                 divs=None,
                 potential_function=None,
                 logger=None,
                 **base_opts
                 ):
        """
        :param base_opts: base opts to use when running
        :type base_opts:
        """
        self.domain = domain
        base_opts['domain'] = domain
        self.divs = divs
        base_opts['divs'] = divs
        self.potential_function = potential_function
        base_opts['potential_function'] = potential_function
        self.opts = base_opts

        if isinstance(logger, Logger):
            self.logger = logger
        elif logger is True:
            self.logger = Logger()
        elif logger is False or logger is None:
            self.logger = NullLogger()
        else:
            self.logger = Logger(logger)

    def __repr__(self):
        if self.potential_function is not None:
            return "{}({}, pts={}, pot={})".format(
                type(self).__name__,
                self.domain,
                self.divs,
                self.potential_function
            )
        else:
            return "{}({}, pts={}, pot={})".format(
                type(self).__name__,
                self.domain,
                self.divs,
                self.potential_function
            )

    def _logger(self, logger):
        return self.logger if logger is None else logger

    @abc.abstractmethod
    def get_grid(self, domain=None, divs=None, **kwargs):
        raise NotImplementedError("abstract interface")
    def grid(self, domain=None, divs=None, **kwargs):
        if domain is None:
            domain = self.domain
        if divs is None:
            divs = self.divs

        if domain is None:
            raise ValueError("need a value for `domain`")
        if divs is None:
            raise ValueError("need a value for `divs`")

        return self.get_grid(domain=domain, divs=divs, **kwargs)

    @abc.abstractmethod
    def get_kinetic_energy(self, grid=None, mass=None, hb=1, **kwargs):
        raise NotImplementedError("abstract interface")
    def handle_kinetic_coupling(self, grid, ke_1D, g, g_deriv, hb=1, logger=None, **kwargs):
        logger = self._logger(logger)
        if g is not None:
            with logger.block(tag="handling kinetic coupling"):
                if g_deriv is None:
                    raise ValueError(
                        "if a function for `g` is supplied, also need a function, `g_deriv` for the second derivative of `g`"
                    )
                # add the average value of `g` across the grid points
                if isinstance(g, (int, float, np.integer, np.floating)):
                    logger.log_print("constant G-matrix element")
                    g_vals = np.full(grid.shape, g)
                    g_deriv_vals = np.zeros(grid.shape)
                else:
                    logger.log_print("variable G-matrix element")
                    try:
                        iter(g)
                    except TypeError:
                        g_vals = g(grid)
                    else:
                        g_vals = np.asanyarray(g)

                    try:
                        iter(g_deriv)
                    except TypeError:
                        g_deriv_vals = g_deriv(grid)
                    else:
                        g_deriv_vals = np.asanyarray(g_deriv)

                g_vals = 1 / 2 * (g_vals[:, np.newaxis] + g_vals[np.newaxis, :])
                g_deriv_vals = (hb ** 2) / 2 * np.diag(g_deriv_vals)
                ke_1D = ke_1D * g_vals + g_deriv_vals
        return ke_1D
    def kinetic_energy(self, grid=None, mass=None, hb=1, g=None, g_deriv=None, **kwargs):

        if grid is None:
            grid = self.grid()

        if g is not None:
            mass = 1

        if mass is None:
            raise ValueError("need a value for the mass")

        ke_1D = self.get_kinetic_energy(grid=grid, mass=mass, g=g, g_deriv=g_deriv, hb=hb, **kwargs)
        ke_1D = self.handle_kinetic_coupling(grid, ke_1D, g, g_deriv, hb=hb, **kwargs)

        return ke_1D

    def real_momentum(self, grid=None, mass=None, hb=1, **kwargs):
        raise NotImplementedError("real momentum needs to be implemented")

    def potential_energy(self,
                         grid=None,
                         potential_function=None,
                         potential_values=None,
                         potential_grid=None,
                         logger=None,
                         **pars
                         ):
        """
        Calculates the potential energy at the grid points based
        on dispatching on the input form of the potential

        :param grid: the grid of points built earlier in the DVR
        :type grid:
        :param potential_function: a function to evaluate the potential energy at the points
        :type potential_function:
        :param potential_values: the values of the potential at the DVR points
        :type potential_values:
        :param potential_grid: a grid of points and values to be interpolated
        :type potential_grid:
        :param pars: ignored keyword arguments
        :type pars:
        :return:
        :rtype:
        """

        if grid is None:
            grid = self.grid()

        if isinstance(grid, tuple) and len(grid) == 2: # for FBR DVRs
            grid, tfs = grid

        if potential_function is None and potential_grid is None and potential_values is None:
            potential_function = self.potential_function

        logger = self._logger(logger)
        if potential_function is not None:
            logger.log_print('evaluating potential function over grid')
            # explicit potential function passed; map over coords
            pf=potential_function

            dim = len(grid.shape)
            if dim > 1:
                npts = np.prod(grid.shape[:-1], dtype=int)
                grid = np.reshape(grid, (npts, grid.shape[-1]))
                pot = sp.diags([pf(grid)], [0])
            else:
                pot = np.diag(pf(grid))
        elif potential_values is not None:
            logger.log_print('constructing potential matrix from diagonal values')
            # array of potential values at coords passed
            dim = len(grid.shape)
            if dim > 1:
                pot = sp.diags([potential_values], [0])
            else:
                pot = np.diag(potential_values)
        elif potential_grid is not None:
            logger.log_print('interpolating potential over grid')
            # TODO: extend to include ND, scipy.griddata

            dim = len(grid.shape)
            if dim > 1:
                dim -= 1
                npts = npts = np.prod(grid.shape[:-1], dtype=int)
                grid = np.reshape(grid, (npts, grid.shape[-1]))

            if dim == 1:
                # use a cubic spline interpolation
                interpolator = lambda g1, g2: interp.interp1d(g1[:, 0], g1[:, 1], kind='cubic')(g2)
            else:
                # use griddata to do a general purpose interpolation
                def interpolator(g, g2):
                    # g is an np.ndarray of potential points and values
                    # g2 is the set of grid points to interpolate them over

                    shape_dim = len(g.shape)
                    if shape_dim == 2:
                        points = g[:, :-1]
                        vals = g[:, -1]
                        return interp.griddata(points, vals, g2)
                    else:
                        # assuming regular structured grid
                        mesh = np.moveaxis(g, 0, shape_dim)
                        points = tuple(np.unique(x) for x in mesh[:-1])
                        vals = mesh[-1]
                        return interp.interpn(points, vals, g2)
            wtf = np.nan_to_num(interpolator(potential_grid, grid))
            pot = sp.diags([wtf], [0])
        else:
            raise DVRException("couldn't construct potential matrix")

        return pot

    def hamiltonian(self, kinetic_energy=None, potential_energy=None, potential_threshold=None, **pars):
        """
        Calculates the total Hamiltonian from the kinetic and potential matrices

        :param kinetic_energy:
        :type kinetic_energy:
        :param potential_energy:
        :type potential_energy: np.ndarray | sp.spmatrix
        :param potential_threshold:
        :type potential_threshold:
        :param pars:
        :type pars:
        :return:
        :rtype:
        """

        if potential_threshold is not None:
            diag = potential_energy.diagonal()
            chops = np.where(diag > 0)
            if len(chops) == 0:
                return kinetic_energy + potential_energy
            chops = chops[0]

            ham = kinetic_energy + potential_energy
            ham[chops, :] = 0
            ham[:, chops] = 0

            return ham

        else:
            return kinetic_energy + potential_energy

    def wavefunctions(self, hamiltonian=None, num_wfns=25, nodeless_ground_state=False, diag_mode=None, logger=None, **pars):
        """
        Calculates the wavefunctions for the given Hamiltonian.
        Doesn't support any kind of pruning based on potential values although that might be a good feature
        to support explicitly in the future

        :param hamiltonian:
        :type hamiltonian:
        :param num_wfns:
        :type num_wfns:
        :param nodeless_ground_state:
        :type nodeless_ground_state:
        :param diag_mode:
        :type diag_mode:
        :param pars:
        :type pars:
        :return:
        :rtype:
        """


        logger = self._logger(logger)
        with logger.block(tag="diagonalizing Hamiltonian"):
            logger.log_print([
                "dimension={dimension}",
                "density={density:.3f}%",
                "mode={mode}"
            ],
                dimension=hamiltonian.shape,
                density=100 * hamiltonian.nnz/np.prod(hamiltonian.shape) if isinstance(hamiltonian, sp.spmatrix) else 100,
                mode=diag_mode
            )
            if isinstance(hamiltonian, sp.spmatrix) and diag_mode == 'dense':
                hamiltonian = hamiltonian.toarray()
            if isinstance(hamiltonian, sp.spmatrix):
                import scipy.sparse.linalg as la
                engs, wfns = la.eigsh(hamiltonian, num_wfns, which='SM')
            else:
                engs, wfns = np.linalg.eigh(hamiltonian)
                if num_wfns is not None:
                    engs = engs[:num_wfns]
                    wfns = wfns[:, :num_wfns]

        if nodeless_ground_state:
            logger.log_print('correcting phases to construct nodeless ground state')
            s = np.sign(wfns[:, 0])
            wfns *= s[:, np.newaxis]
        return engs, wfns

    def run(self, result='wavefunctions', logger=None, grid=None, potential_energy=None, kinetic_energy=None, hamiltonian=None, **opts):
        """
        :return:
        :rtype: DVRResults
        """
        from .Wavefunctions import DVRWavefunctions

        opts = dict(self.opts, **opts)

        try:
            self._opts = self.opts
            self.opts = opts

            logger = self._logger(logger)
            with logger.block(tag="Running DVR"):
                logger.log_print("{dvr}", dvr=self)
                logger.log_print(opts, message_prepper=logger.prep_dict)
                opts['logger'] = logger

                res = DVRResults(parent=self, **self.opts)

                with logger.block(tag="constructing grid"):
                    if grid is None:
                        grid = self.grid(**self.opts)
                    res.grid = grid
                    if result == 'grid':
                        return res


                with logger.block(tag="constructing potential matrix"):
                    if potential_energy is None:
                        potential_energy = self.potential_energy(grid=res.grid, **opts)
                    res.potential_energy = potential_energy
                    if result == 'potential_energy':
                        return res


                with logger.block(tag="constructing kinetic matrix"):
                    if kinetic_energy is None:
                        kinetic_energy = self.kinetic_energy(grid=res.grid, **opts)
                    res.kinetic_energy = kinetic_energy
                    if result == 'kinetic_energy':
                        return res


                with logger.block(tag="building Hamiltonian"):
                    if hamiltonian is None:
                        hamiltonian = self.hamiltonian(
                            kinetic_energy=res.kinetic_energy,
                            potential_energy=res.potential_energy,
                            **opts
                        )
                    res.hamiltonian = hamiltonian
                    if result == 'hamiltonian':
                        return res


                with logger.block(tag="evaluating wavefunctions"):
                    energies, wfn_data = self.wavefunctions(
                        hamiltonian=res.hamiltonian,
                        **opts
                    )
                    wfns = DVRWavefunctions(energies=energies, wavefunctions=wfn_data, grid=res.grid, results=res, **opts)
                    res.wavefunctions = wfns

                return res
        finally:
            self.opts = self._opts

class DVRResults:
    """
    A subclass that can wrap all of the DVR run parameters and results into a clean interface for reuse and extension
    """
    def __init__(self,
                 grid=None,
                 kinetic_energy=None,
                 potential_energy=None,
                 hamiltonian=None,
                 wavefunctions=None,
                 parent=None,
                 **opts
                 ):

        # self.parent = None
        self.grid = grid
        self.kinetic_energy = kinetic_energy
        self.potential_energy = potential_energy
        self.parent = parent
        self.wavefunctions = wavefunctions
        self.hamiltonian = hamiltonian
        self.opts = opts

    @property
    def dimension(self):
        dim = len(self.grid.shape)
        if dim > 1:
            dim -= 1
        return dim

    def plot_potential(self, plot_class=None, figure=None, plot_units=None, energy_threshold=None, zero_shift=False, **opts):
        """
        Simple plotting function for the potential.
        Should be updated to deal with higher dimensional cases

        :param plot_class: the graphics class to use for the plot
        :type plot_class: McUtils.Plots.Graphics
        :param opts: plot styling options
        :type opts:
        :return:
        :rtype: McUtils.Plots.Graphics
        """
        from McUtils.Plots import Plot, ContourPlot

        # get the grid for plotting
        MEHSH = self.grid
        dim = self.dimension
        if dim == 1:
            mesh = [MEHSH]
        else:
            mesh = np.moveaxis(MEHSH, dim, 0)
        if plot_class is None:
            if dim == 1:
                plot_class = Plot
            elif dim == 2:
                plot_class = ContourPlot
            else:
                raise DVRException("{}.{}: don't know how to plot {} dimensional potential".format(
                    type(self).__name__,
                    'plot',
                    dim
                ))

        pot = self.potential_energy.diagonal()
        if isinstance(plot_units, str) and plot_units == 'wavenumbers':
            pot = pot * UnitsData.convert("Hartrees", "Wavenumbers")

        if zero_shift:
            pot = pot - np.min(pot)

        if energy_threshold:
            pot = pot.copy()
            pot[pot > energy_threshold] = energy_threshold

        return plot_class(*mesh, pot.reshape(mesh[0].shape), figure=figure, **opts)

class DVRException(Exception):
    """
    Base exception class for working with DVRs
    """





"""
Provides an abstract Hamiltonian object that can be used when building representations
"""

__all__ = [
    "Representation",
    "ExpansionRepresentation"
]

import itertools
import tracemalloc

import numpy as np, itertools as ip, scipy.sparse as sp, time, gc, tempfile as tf

from McUtils.Numputils import SparseArray
import McUtils.Numputils as nput
from McUtils.Scaffolding import Logger, NullLogger

from .Operators import Operator
from .StateSpaces import BraKetSpace, SelectionRuleStateSpace, StateSpaceMatrix

# for interactive work
__reload_hook__ = [ '.StateSpaces', ".Operators" ]

class Representation:
    """
    A `Representation` provides a simple interface to build matrix representations of operators expressed
    in high-dimensional spaces.

    """

    def __init__(self, compute, basis, name=None, logger=None,
                 selection_rules=None,
                 selection_rule_steps=None,
                 memory_constrained=False
                 ):
        """
        :param compute: the function that turns indices into values
        :type compute: callable | Operator
        :param basis: the basis quanta used in the representations
        :type basis: RepresentationBasis
        :param logger: logger for printing out debug info
        :type logger: None | Logger
        """
        if isinstance(compute, Operator):
            operator = compute
            compute = self._compute_op_els
        else:
            operator = None
        self.operator = operator
        self._compute = compute
        self.basis = basis
        self.dims = basis.dimensions
        self._diminds = None
        if logger is None:
            logger = NullLogger()
        self.logger = logger
        self.array = StateSpaceMatrix(basis)
        self._selection_rules = selection_rules
        self._selection_rule_steps = selection_rule_steps
        self.name=name
        self.memory_constrained = memory_constrained

    @property
    def parallelizer(self):
        if self.operator is not None:
            return self.operator.parallelizer
        else:
            return None

    def compute(self, inds, **kwargs):
        # if isinstance(inds, BraKetSpace):
        #     # allows us to use cached stuff
        #     return self.array.compute_values(self._compute, inds)
        # else:
        return self._compute(inds, **kwargs)
    def compute_cached(self, inds):
        if isinstance(inds, BraKetSpace):
            # allows us to use cached stuff
            self.array._compute_uncached_values(self._compute, inds)
        else:
            raise ValueError("Can only compute cached values when given explicit BraKets")
    @property
    def chunk_size(self):
        if self.operator is None:
            return None
        else:
            return self.operator.chunk_size

    def clear_cache(self):
        # print(">>>>>>>>>>>>> wat", self.compute, self.operator)
        if self.operator is not None:
            # print("?????")
            self.operator.clear_cache()
        elif hasattr(self.compute, 'clear_cache'):
            self.compute.clear_cache()
        gc.collect()
    def _compute_op_els(self, inds, check_orthogonality=True, memory_constrained=None):
        res = self.operator.get_elements(inds,
                                          check_orthogonality=check_orthogonality,
                                          memory_constrained=self.memory_constrained if memory_constrained is None else memory_constrained
                                          )
        return res

    @property
    def diag(self):
        return self[self.dim_inds, self.dim_inds]
    @property
    def ndims(self):
        return int(np.prod(self.dims))
    @property
    def dim_inds(self):
        # we probably want to avoid using this...
        if self._diminds is None:
            self._diminds = np.arange(self.ndims)
        return self._diminds
    def _get_inds(self, n):
        if isinstance(n, slice):
            start = n.start
            stop = n.stop
            step = n.step
            if start is None:
                start = 0
            elif start < 0:
                start = self.ndims + start
            if stop is None:
                stop = self.ndims
            elif stop < 0:
                stop = self.ndims + stop
            if step is None:
                step = 1
            return np.arange(start, stop, step)
        elif isinstance(n, np.ndarray):
            negs = np.where(n < 0)
            if len(negs[0]) > 0:
                n[negs] += self.ndims
            return n
    # @profile
    def get_brakets(self, states, check_orthogonality=True, memory_constrained=False):
        """
        Computes term elements based on getting a BraKetSpace.
        Can directly pass element specs through, since the shape management shit
        is managed by the BraKetSpace

        :param states:
        :type states: BraKetSpace | Tuple[np.ndarray, np.ndarray]
        :return:
        :rtype:
        """
        if not isinstance(states, BraKetSpace):
            states = BraKetSpace.from_indices(states, basis=self.basis)

        self.logger.log_print(
            "evaluating in BraKet space {states}",
            states=states
        )

        return self.compute(states, check_orthogonality=check_orthogonality, memory_constrained=memory_constrained)

    def get_element(self, n, m):
        """
        Computes term elements.
        Determines first whether it needs to pull single elements or blocks of them.

        :param n:
        :type n:
        :param m:
        :type m:
        :return:
        :rtype:
        """

        dims = self.dims
        # ndims = int(np.prod(dims))
        idx = (n, m)

        # There are two possible modes for this pulling individual elements or pulling blocks
        # the block pulling can be quite a bit faster, so we try to detect first off if we want to do that
        pull_elements = True
        # first we check if we've got something like from `np.ix_(n, m)`
        if isinstance(n, np.ndarray) and isinstance(m, np.ndarray):
            if len(n.shape) > 1 and len(m.shape) > 1:
                pull_elements = False
        if pull_elements:
            # next we check to see if idx is really just a single element
            pull_elements = all(isinstance(x, (int, np.integer)) for x in idx)
            if not pull_elements:
                # if not a single element, we make sure there are no slices
                pull_elements = all(not isinstance(x, (int, np.integer, slice)) for x in idx)
                # print(">?>", pull_elements, idx)
                if pull_elements:
                    e1 = len(idx[0])
                    pull_elements = all(len(x) == e1 for x in idx)
                    # print(">??", pull_elements)

        # This manages figuring out which indices we pull...

        # We figure out the row spec
        if not isinstance(n, int):
            if isinstance(n, np.ndarray):
                n = n.flatten()
            if not isinstance(n, slice):
                n = np.array(n, dtype=int)
            n = self._get_inds(n)
        else:
            n = [n]
        # Then the column spec
        if not isinstance(m, int):
            if isinstance(m, np.ndarray):
                m = m.flatten()
            if not isinstance(m, slice):
                m = np.array(m)
            m = self._get_inds(m)
        else:
            m = [m]

        if pull_elements:
            # If we're just pulling elements we need only unravel those indices
            n = np.unravel_index(n, dims)
            m = np.unravel_index(m, dims)
        else:
            # If we're pulling blocks we need to compute the product of the row
            #  and column indices to get the total index spec
            blocks = np.array(list(ip.product(n, m)))
            n = np.unravel_index(blocks[:, 0], dims)
            m = np.unravel_index(blocks[:, 1], dims)

        # we define a temporary helper to pad repeat the indices if necessary
        def pad_lens(a, b):
            if isinstance(a, (int, np.integer)) and not isinstance(b, (int, np.integer)):
                a = np.full((len(b),), a)
            if isinstance(b, (int, np.integer)) and not isinstance(a, (int, np.integer)):
                b = np.full((len(a),), b)
            return a, b

        i = tuple(pad_lens(a, b) for a, b in zip(n, m))
        els = self.compute(i)
        if isinstance(els, int) and els == 0:
            # short-circuited :|
            return els

        elif not pull_elements:
            shp = (len(np.unique(blocks[:, 0])), len(np.unique(blocks[:, 1])))
            extra_shp = els.shape[:-1] # some terms will return higher-dimensional results?
            # for sparse arrays this happens in-place :|
            els = els.reshape(extra_shp + shp).squeeze()

        return els

    def __getitem__(self, item):
        if isinstance(item, BraKetSpace): # short circuit here
            return self.get_brakets(item)

        if not isinstance(item, tuple):
            item = (item,)
        if len(item) == 1:
            item = item + (slice(None, None, None),)
        if len(item) > 2:
            raise Exception("index spec '{}' must be of dimension 2".format(item))
        else:
            inds = np.array(item)
            if inds.dtype == np.dtype(int):
                return self.get_brakets(inds)
            else:
                return self.get_element(*item)

    def __rmul__(self, other):
        if isinstance(other, (int, float, np.integer, np.floating)):
            return ExpansionRepresentation([other], [self], self.basis,
                                           name=self.name,
                                           logger=self.logger,
                                           memory_constrained=self.memory_constrained
            )
        else:
            raise TypeError("operator * not defined for objects of type {0} and {1} (only numbers are supported with {0})".format(
                type(self).__name__,
                type(other).__name__
            ))
    def __mul__(self, other):
        if isinstance(other, (int, float, np.integer, np.floating)):
            return ExpansionRepresentation([other], [self],
                                           self.basis,
                                           name=self.name,
                                           logger=self.logger,
                                           memory_constrained=self.memory_constrained
                                           )
        else:
            raise TypeError("operator * not defined for objects of type {0} and {1} (only numbers are supported with {0})".format(
                type(self).__name__,
                type(other).__name__
            ))
    def __add__(self, other):
        if isinstance(other, Representation):
            if other.dims != self.dims:
                if isinstance(other, ExpansionRepresentation):
                    return other + self
                raise ValueError("Can't combine TermComputer objects with dim {} and {}".format(
                    self.dims,
                    other.dims
                ))
            return ExpansionRepresentation([1, 1],
                                           [self, other],
                                           self.basis,
                                           name=self.name + other.name if self.name is not None and other.name is not None else None,
                                           logger=self.logger,
                                           memory_constrained=self.memory_constrained
            )
        else:
            raise TypeError("operator + not defined for objects of type {0} and {1} (only subclasses of TermComputer are supported with {0})".format(
                type(self).__name__,
                type(other).__name__
            ))

    @staticmethod
    def _get_dim_string(dims):
        if len(dims) < 7:
            return ", ".join(str(s) for s in dims)
        else:
            return "{}, ({} more), {}".format(
                ", ".join(str(s) for s in dims[:2]),
                len(dims) - 4,
                ", ".join(str(s) for s in dims[-2:])
            )

    def __repr__(self):
        if self.name is None:
            return "{}(<{}>, {})".format(
                type(self).__name__,
                self._get_dim_string(self.dims) if hasattr(self, 'dims') else '???',
                self.operator if self.operator is not None else self.compute
            )
        else:
            return "{}<{}>".format(
                type(self).__name__, self.name
            )

    @property
    def selection_rules(self):
        """

        :return:
        :rtype:
        """
        if self._selection_rules is None:
            if self.operator is not None:
                return self.operator.selection_rules

            else:
                return None
        else:
            return self._selection_rules
    @selection_rules.setter
    def selection_rules(self, rules):
        if isinstance(rules[0], (int, np.integer)):
            raise ValueError('selection rules expected to be a list of lists')
        self._selection_rules = rules

    @property
    def selection_rule_steps(self):
        """
        :return:
        :rtype:
        """
        if self._selection_rule_steps is None:
            if self.operator is not None:
                return self.operator.selection_rule_steps

            else:
                return None
        else:
            return self._selection_rule_steps

    @property
    def is_diagonal(self):
        if self.operator is not None:
            return self.operator.is_diagonal
        else:
            return False
    @property
    def is_zero(self):
        if self.operator is not None:
            return self.operator.is_zero
        else:
            return False

    @property
    def skipped_indices(self):
        """

        :return:
        :rtype:
        """
        if self.operator is not None:
            return self.operator.skipped_indices
        else:
            return None

    def get_transformed_space(self, space, parallelizer=None, logger=None, **opts):
        """
        Returns the state space obtained by using the
        held operator to transform `space`

        :param space:
        :type space:
        :return: connected state spaces
        :rtype: SelectionRuleStateSpace
        """

        if parallelizer is None:
            parallelizer = self.parallelizer

        if self.operator is not None:
            return self.operator.get_transformed_space(space, rules=self.selection_rules, parallelizer=parallelizer, logger=logger,
                                                       **opts
                                                       )
        elif self.selection_rules is not None:
            new = space.apply_selection_rules(self.selection_rules, parallelizer=parallelizer, logger=logger,
                                               **opts
                                               )
            skips = Operator._get_skip_trans(self.skipped_indices, space.ndim)
            if skips is not None:
                new = new.filter_transitions(skips)
            return new
        else:
            raise ValueError("can't get a transformed space without a held operator or selection rules")


    def apply(self, other):
        raise NotImplementedError("This code path is temporarily abandoned")

        if self.operator is None:
            raise ValueError("")

        if not isinstance(other, (Representation, StateSpaceMatrix)):
            raise TypeError("{} doesn't support application to objects that don't provide a state space".format(
                type(self).__name__
            ))

        if isinstance(other, Representation):
            other = other.array

        other_states = other.brakets.bras
        new_states = self.operator.get_transformed_space(other_states)
        # try:
        brakets = new_states.get_representation_brakets()
        # raise Exception(brakets.bras.excitations[:5], brakets.kets.excitations[:5])
        # except:
        #     raise Exception(other_states, other)

        self.compute_cached(brakets)

        return self.array.dot(other)

    # from memory_profiler import profile
    def _build_rep_matrix(self, row_inds, col_inds, N, sub, method='unique'):

        if method == 'low_mem':
            # logger.log_print("building upper/triangle indices...")
            # upper triangle of indices
            utri_pos = np.where(row_inds < col_inds)[0]
            diag_pos = np.where(row_inds == col_inds)
            if len(diag_pos) > 0:
                diag_pos = diag_pos[0]

            utri_rows = row_inds[utri_pos]
            utri_cols = col_inds[utri_pos]
            if len(utri_pos) + len(diag_pos) < len(row_inds):
                ltri_pos = np.setdiff1d(np.arange(len(row_inds)),
                                        np.concatenate([utri_pos, diag_pos])
                                        )

                # ltri_rows = row_inds[ltri_pos]

                # this was originally just np.unique but the memory
                # for constructing the sort arrays got large?
                # remove dupes by focusing on col inds first
                ltri_cols = col_inds[ltri_pos]
                col_grps = nput.group_by(ltri_pos, ltri_cols)[0]
                unduped_pos = []
                utri_rem_rows = np.full(len(utri_rows), True) # we build a mask to cut down on duplicate tests
                for k,idx in zip(*col_grps):
                    possible_matches = np.where(utri_rows[utri_rem_rows] == k)
                    if len(possible_matches) > 0:
                        possible_matches = possible_matches[0]
                        # now we match corresponding upper-triangle cols against lower-tri rows
                        utri_test_cols = utri_cols[utri_rem_rows][possible_matches]
                        # we mask out the matched positions so we don't test
                        # them on later iterations
                        utri_rem_rows[utri_rem_rows][possible_matches] = False

                        # we assume no duplicates at this point
                        # so we just test the utri_test_cols against
                        # the different column vals to see if we have
                        # any dupes

                        good_spots = np.where(
                            np.all(
                                utri_test_cols[np.newaxis, :] != row_inds[idx][:, np.newaxis],
                                axis=1
                                )
                        )
                        # print(good_spots.shape, len(idx))
                        # print(idx)
                        if len(good_spots) > 0:
                            good_spots = good_spots[0]
                            unduped_pos.append(idx[good_spots])
                    else: # no upper-triangle rows matched this lower-triangle column
                        unduped_pos.append(idx)

                ltri_pos = np.concatenate(unduped_pos)

                tri_pos = np.concatenate([utri_pos, ltri_pos], axis=0)

            else:
                tri_pos = utri_pos

            tri_rows = row_inds[tri_pos]
            tri_cols = col_inds[tri_pos]
            tri_vals = sub[tri_pos]

            up_tri = np.array([tri_rows, tri_cols]).T
            lo_tri = np.array([tri_cols, tri_rows]).T

            if len(diag_pos) > 0: # diags can get duped...somehow
                _, wtf_pos = np.unique(row_inds[diag_pos], return_index=True)
                diag_pos = diag_pos[np.sort(wtf_pos)]

            diag_inds = row_inds[diag_pos]
            diag_vals = sub[diag_pos]
            diag = np.array([diag_inds, diag_inds]).T

            full_inds = np.concatenate([up_tri, lo_tri, diag], axis=0)
            full_dat = np.concatenate([tri_vals, tri_vals, diag_vals], axis=0)

        elif method == 'unique':

            # upper triangle of indices
            up_tri = np.array([row_inds, col_inds]).T
            # lower triangle is made by transposition
            low_tri = np.array([col_inds, row_inds]).T
            # but now we need to remove the duplicates, because many sparse matrix implementations
            # will sum up any repeated elements...
            full_inds = np.concatenate([up_tri, low_tri])
            full_dat = np.concatenate([sub, sub], axis=0)

            _, idx = np.unique(full_inds, axis=0, return_index=True)
            sidx = np.sort(idx)
            full_inds = full_inds[sidx]
            full_dat = full_dat[sidx]

        else:
            raise ValueError('bad method {}'.format(method))

        # N = len(total_space)
        if full_dat.ndim > 1:
            # need to tile the inds enough times to cover the shape of full_dat
            # _, inds = np.unique(full_inds, axis=0, return_index=True)
            # sidx = np.sort(inds)
            # full_inds = full_inds[sidx]
            # full_dat = full_dat[sidx]

            ext_shape = full_dat.shape[1:]
            shape = (N, N) + ext_shape
            full_dat = np.moveaxis(full_dat, 0, -1).flatten()

            rows, cols = full_inds.T
            full_inds = np.empty(
                (2 + len(ext_shape), len(full_dat)),
                dtype=int
            )
            block_size = len(rows)
            # create a tensor of indices that will be used to appropriately tile
            # the system
            ind_tensor = np.moveaxis(np.indices(ext_shape), 0, -1).reshape(-1, len(ext_shape))
            for n, idx in enumerate(ind_tensor):
                s, e = n * block_size, (n + 1) * block_size
                full_inds[0, s:e] = rows
                full_inds[1, s:e] = cols
                for j, x in enumerate(idx):
                    full_inds[2 + j, s:e] = x
        else:
            full_inds = full_inds.T
            shape = (N, N)

        return SparseArray.from_data((full_dat, full_inds), shape=shape, cache_block_data=False)

    # from memory_profiler import profile
    # @profile
    def get_representation_matrix(self,
                                  coupled_space,
                                  total_space,
                                  filter_space=None,
                                  diagonal=False,
                                  logger=None,
                                  zero_element_warning=True,
                                  clear_sparse_caches=True,
                                  clear_operator_caches=True,
                                  assume_symmetric=True,
                                  remove_duplicates=True,
                                  memory_constrained=None
                                  ):
        """
        Actively constructs a perturbation theory Hamiltonian representation

        :param h:
        :type h:
        :param cs:
        :type cs:
        :return:
        :rtype:
        """

        if not assume_symmetric:
            raise NotImplementedError("don't have non-symmetric representations implemented")

        if memory_constrained is None:
            memory_constrained = self.memory_constrained

        if logger is None:
            logger = self.logger

        if not isinstance(coupled_space, BraKetSpace):
            if diagonal:
                m_pairs = BraKetSpace(coupled_space, coupled_space)
            else:
                m_pairs = coupled_space.get_representation_brakets(other=filter_space)
        else:
            m_pairs = coupled_space

        if remove_duplicates:
            m_pairs = m_pairs.remove_duplicates(assume_symmetric=assume_symmetric)

        if len(m_pairs) > 0:
            # logger.log_print(["coupled space dimension {d}"], d=len(m_pairs))
            sub = self.get_brakets(m_pairs, check_orthogonality=not diagonal, memory_constrained=memory_constrained)
            if isinstance(sub, SparseArray):
                sub = sub.asarray()
            if clear_sparse_caches:
                SparseArray.clear_cache()
            if clear_operator_caches:
                self.clear_cache()
        else:
            logger.log_print('no states to couple!', log_level=logger.LogLevel.Debug)
            sub = 0

        logger.log_print("constructing sparse representation...", log_level=logger.LogLevel.Debug)

        N = len(total_space)
        if isinstance(sub, (int, np.integer, np.floating, float)):
            if sub == 0:
                sub = SparseArray.empty((N, N), dtype=float)
            else:
                raise ValueError("Using a constant shift of {} will force representation to be dense...".format(sub))
                sub = np.full((N, N), sub)
        else:
            if zero_element_warning:
                zeros = np.where(sub == 0.)
                if len(zeros) > 0 and len(zeros[0]) > 0:
                    zeros = zeros[0]
                    bad_pairs = (m_pairs.bras.take_subspace(zeros), m_pairs.kets.take_subspace(zeros))
                    raise ValueError(
                        ('got {0} zero elements from {1}; if you expect zeros, set `zero_element_warning=False`. '
                         'First zero element: <|{2}|{1}|{3}>').format(len(zeros), self,
                                                                      bad_pairs[0].excitations[0],
                                                                      bad_pairs[1].excitations[0]
                                                                      )
                    )
            if diagonal and coupled_space is total_space: # fast shortcut
                sub = SparseArray.from_diag(sub)
            else:
                # we assume duplicates were removed earlier

                utri, full_inds = self._pull_rep_mat_inds(logger, total_space, m_pairs)

                logger.log_print("getting vals array...", log_level=logger.LogLevel.Debug)
                if len(sub) != len(m_pairs): # happens when multioperator asks for just one state
                    if len(m_pairs) > 1:
                        raise ValueError("matrix element evaluation broke in way I don't understand")
                    sub = sub[np.newaxis]

                full_dat = np.concatenate([sub, sub[utri]], axis=0)
                del sub, utri

                sorting = np.argsort(full_inds[:, 0])
                np.take(full_dat, sorting, out=full_dat, axis=0, mode='clip')
                np.take(full_inds, sorting, out=full_inds, axis=0, mode='clip')
                del sorting

                full_dat, full_inds, shape = self._handle_multidim_rep_mat(logger, full_dat, full_inds, N)

                logger.log_print("building sparse tensor...", log_level=logger.LogLevel.Debug)
                # another attempt
                # if False and memory_constrained:
                #     with tf.NamedTemporaryFile() as full_dat_dump:
                #         with tf.NamedTemporaryFile() as full_inds_dump:
                #             full_dat_map = np.memmap(full_dat_dump.name, dtype=full_dat.dtype, shape=full_dat.shape)
                #             full_dat_map[:] = full_dat[:]
                #             del full_dat
                #
                #             full_inds_map = np.memmap(full_inds_dump.name, dtype=full_inds.dtype, shape=full_inds.shape)
                #             full_inds_map[:] = full_inds[:]
                #             del full_inds
                #
                #             sub = SparseArray.from_data((full_dat_map, full_inds_map),
                #                                         shape=shape,
                #                                         cache_block_data=False,
                #                                         logger=logger
                #                                         # layout='coo'
                #                                         )
                #             # sub2 = SparseArray.from_data((full_dat, full_inds), shape=shape, cache_block_data=False, logger=logger)
                #             #
                #             # raise Exception(np.sum(sub.asarray() - sub2.asarray()))
                #
                # else:
                cs = self.chunk_size
                if cs is not None:
                    cs = int(cs)
                    blocks = int(np.ceil(len(full_dat) / cs))
                    sub = None
                    for i in range(blocks):
                        s = i*cs; e = min((i+1)*cs, len(full_dat))
                        init = SparseArray.initializer_list(full_dat[s:e], full_inds[:, s:e])
                        subub = SparseArray.from_data(init,
                                                shape=shape,
                                                cache_block_data=False,
                                                logger=logger
                                                , init_kwargs=dict(assume_sorted=True)
                                                )
                        if sub is None:
                            sub = subub
                        else:
                            sub += subub

                    del full_dat
                    del full_inds

                else:
                    init = SparseArray.initializer_list(full_dat, full_inds)
                    del full_dat
                    del full_inds
                    sub = SparseArray.from_data(init,
                                                shape=shape,
                                                cache_block_data=False,
                                                logger=logger
                                                , init_kwargs=dict(assume_sorted=True)
                                                )
                    del init


        return sub

    # @profile
    def _pull_rep_mat_inds(self, logger, total_space, m_pairs):
        logger.log_print("finding row/column indices...", log_level=logger.LogLevel.Debug)
        # figure out the appropriate inds for this data in the sparse representation

        # I should probably add some level of chunking on this?
        # or force `bras` to keep its data on disk?
        row_inds = total_space.find(m_pairs.bras, minimal_dtype=True)
        col_inds = total_space.find(m_pairs.kets, minimal_dtype=True)

        logger.log_print("constructing full row/col index array...", log_level=logger.LogLevel.Debug)
        # ninds = len(row_inds)
        # first_bits = sidx[sidx < ninds]
        # last_bits = sidx[sidx >= ninds] - ninds

        utri = row_inds != col_inds
        full_inds = np.array([
            np.concatenate([row_inds, col_inds[utri]]),
            np.concatenate([col_inds, row_inds[utri]])
        ]).T
        # _, idx = np.unique(full_inds, axis=0, return_index=True)
        # sidx = np.sort(idx)

        # nddup = len(np.unique(diag_inds))
        # logger.log_print("pre/pos: {nfull}/{nun} diags/undiags:{ndig}/{nddup} ??? {ndiff}",
        #                  nfull=len(full_inds), nun=len(sidx),
        #                  ndig=len(diag_inds), nddup=nddup,
        #                  ndiff = len(full_inds) - nddup
        #                  )
        # full_inds = full_inds[sidx]

        return utri, full_inds

    # @profile
    def _handle_multidim_rep_mat(self, logger, full_dat, full_inds, N):

        if full_dat.ndim > 1:
            # need to tile the inds enough times to cover the shape of full_dat

            logger.log_print("making data multi-dim...", log_level=logger.LogLevel.Debug)
            ext_shape = full_dat.shape[1:]
            shape = (N, N) + ext_shape
            full_dat = full_dat.flatten()

            logger.log_print("populating index array...", log_level=logger.LogLevel.Debug)
            rows, cols = full_inds.T
            full_inds = np.empty(
                (len(shape), len(full_dat)),
                dtype=nput.infer_inds_dtype(len(full_dat))
            )

            # create a tensor of indices that will be used to appropriately tile
            # the system
            stride = len(shape)
            # print(len(shape), len(full_dat)/len(rows))
            # row_repeats = np.broadcast_to(rows[:, np.newaxis], (len(rows), len(ext_shape))).flatten()
            # del rows
            # col_repeats = np.broadcast_to(cols[:, np.newaxis], (len(cols), len(ext_shape))).flatten()
            # del cols
            ind_tensor = np.moveaxis(np.indices(ext_shape), 0, -1).reshape(-1, len(ext_shape))
            for n, idx in enumerate(ind_tensor):
                # we want this to remain sorted, so
                # we want to tile this in such a way that
                # all of the things associated with row 0 come first
                # then with row 1
                # then row 2, etc
                # To do so, we fill with an offset of num extra columns
                full_inds[0, n::stride] = rows
                full_inds[1, n::stride] = cols
                for j, x in enumerate(idx):
                    full_inds[2 + j, n::stride] = x
            del rows
            del cols

        else:
            full_inds = full_inds.T
            shape = (N, N)

        return full_dat, full_inds, shape

    def get_diagonal_representation(self,
                                  coupled_space,
                                  total_space,
                                  logger=None,
                                  zero_element_warning=True,
                                  clear_sparse_caches=True
                                  ):
        """
        Actively constructs a perturbation theory Hamiltonian representation

        :param h:
        :type h:
        :param cs:
        :type cs:
        :return:
        :rtype:
        """

        if logger is None:
            logger = self.logger
        m_pairs = coupled_space.get_representation_brakets()

        if len(m_pairs) > 0:
            logger.log_print(["coupled space dimension {d}"], d=len(m_pairs), log_level=logger.LogLevel.Debug)
            sub = self[m_pairs]
            if clear_sparse_caches:
                SparseArray.clear_cache()
        else:
            logger.log_print('no states to couple!', log_level=logger.LogLevel.Debug)
            sub = 0

        logger.log_print("constructing sparse representation...", log_level=logger.LogLevel.Debug)

        N = len(total_space)
        if isinstance(sub, (int, np.integer, np.floating, float)):
            if sub == 0:
                sub = SparseArray.empty((N, N), dtype=float)
            else:
                raise ValueError("Using a constant shift of {} will force representation to be dense...".format(sub))
                sub = np.full((N, N), sub)
        else:
            # figure out the appropriate inds for this data in the sparse representation
            row_inds = total_space.find(m_pairs.bras)
            col_inds = total_space.find(m_pairs.kets)

            if zero_element_warning:
                zeros = np.where(sub == 0.)
                if len(zeros) > 0 and len(zeros[0]) > 0:
                    zeros = zeros[0]
                    bad_pairs = (m_pairs.bras.take_subspace(zeros), m_pairs.kets.take_subspace(zeros))
                    raise ValueError(
                        ('got zero elements from {0}; if you expect zeros, set `zero_element_warning=False`. '
                         'First zero element: <|{1}|{0}|{2}>').format(self,
                                                                      bad_pairs[0].excitations[0],
                                                                      bad_pairs[1].excitations[0]
                                                                      )
                    )

            # upper triangle of indices
            up_tri = np.array([row_inds, col_inds]).T
            # lower triangle is made by transposition
            low_tri = np.array([col_inds, row_inds]).T
            # but now we need to remove the duplicates, because many sparse matrix implementations
            # will sum up any repeated elements
            full_inds = np.concatenate([up_tri, low_tri])
            full_dat = np.concatenate([sub, sub])

            _, idx = np.unique(full_inds, axis=0, return_index=True)
            sidx = np.sort(idx)
            full_inds = full_inds[sidx]
            full_dat = full_dat[sidx]
            sub = SparseArray.from_data((full_dat, full_inds.T), shape=(N, N))

        return sub

class ExpansionRepresentation(Representation):
    """
    Provides support for terms that look like `1/2 pGp + 1/2 dV/dQdQ QQ` by computing each term on its own
    """
    def __init__(self, coeffs, computers, basis, name=None, logger=None, memory_constrained=False):
        """
        :param coeffs: The expansion coefficients
        :type coeffs: Iterable[float]
        :param compute: the functions that turns indices into values
        :type compute: Iterable[callable | Operator]
        :param n_quanta: the total quanta used in the representations (necessary for shape reasons)
        :type n_quanta: tuple[int]
        """
        self.coeffs = np.array(coeffs)
        self.computers = [Representation(c, basis) if not isinstance(c, Representation) else c for c in computers]
        super().__init__(None, basis, name=name, logger=logger, memory_constrained=memory_constrained)

    @property
    def is_diagonal(self):
        if self.operator is not None:
            return self.operator.is_diagonal
        else:
            return all(c.is_diagonal for c in self.computers)
    @property
    def is_zero(self):
        if self.operator is not None:
            return self.operator.is_zero
        else:
            return all(c.is_zero for c in self.computers)

    @property
    def skipped_indices(self):
        """
        :return:
        :rtype:
        """
        if self.operator is not None:
            return self.operator.is_diagonal
        else:
            skip_sets = [c.skipped_indices for c,s in zip(self.computers, self.coeffs) if (
                    not c.is_diagonal and
                    not c.is_zero and
                    s != 0
            )]

            if len(skip_sets) > 0 and all(skips is not None and len(skips) > 0 for skips in skip_sets):
                # raise Exception(skip_sets)
                skips = list(skip_sets[0])
                for s in skip_sets[1:]:
                    skips, _, _ = nput.intersection(skips, list(s))
                # raise Exception(skips)
            else:
                skips = None
            return skips

    def clear_cache(self):
        for c in self.computers:
            c.clear_cache()

    def __rmul__(self, other):
        if isinstance(other, (int, float, np.integer, np.floating)):
            return type(self)(self.coeffs * other, self.computers, self.basis,
                              logger=self.logger
                              )
        else:
            raise TypeError(
                "operator * not defined for objects of type {0} and {1} (only numbers are supported with {0})".format(
                    type(self).__name__,
                    type(other).__name__
                ))
    def __mul__(self, other):
        if isinstance(other, (int, float, np.integer, np.floating)):
            return type(self)(self.coeffs*other, self.computers, self.basis,
                              name=self.name,
                              logger=self.logger
                              )
        else:
            raise TypeError(
                "operator * not defined for objects of type {0} and {1} (only numbers are supported with {0})".format(
                    type(self).__name__,
                    type(other).__name__
                ))
    def __add__(self, other):
        if isinstance(other, Representation):
            if other.dims != self.dims:
                raise ValueError("Can't combine TermComputer objects with dim {} and {}".format(
                    self.dims,
                    other.dims
                ))
            if isinstance(other, ExpansionRepresentation):
                return type(self)(
                    np.concatenate([self.coeffs, other.coeffs]),
                    self.computers + other.computers,
                    self.basis,
                    name=self.name + other.name if (self.name is not None and other.name is not None) else None,
                    logger=self.logger
                )
            elif isinstance(other, Representation):
                return type(self)(
                    np.concatenate([self.coeffs, [1]]),
                    self.computers + [other],
                    self.basis,
                    name=self.name + other.name if (self.name is not None and other.name is not None) else None,
                    logger=self.logger
                )
            else:
                raise TypeError(
                    "operator + not defined for objects of type {0} and {1} (only subclasses of TermComputer are supported with {0})".format(
                        type(self).__name__,
                        type(other).__name__
                    ))

        else:
            raise TypeError(
                "operator + not defined for objects of type {0} and {1} (only subclasses of TermComputer are supported with {0})".format(
                    type(self).__name__,
                    type(other).__name__
                ))

    def _dispatch_over_expansion(self, attr, *args, **kwargs):
        els = None
        for c, t in zip(self.coeffs, self.computers):
            if not (isinstance(c, (int, float, np.integer, np.floating)) and c == 0):
                with self.logger.block(tag="in {}".format(t)):
                    start = time.time()
                    bits = getattr(t, attr)(*args, **kwargs)
                    scaled = bits * c
                    if self.ndims == 4:
                        raise ValueError("woop")
                    if isinstance(scaled, (int, float, np.integer, np.floating)) and scaled != 0:
                        # raise Exception(bits, c, scaled, n,m, t)
                        raise ValueError(" ".join([
                            "Adding a constant ({}) to a sparse operator ({}) would cast to dense.",
                            "Error likely occurred in getting elements for {}.",
                            "Explicitly subclass {} if you truly need the constant shift.",
                        ]).format(
                            scaled,
                            els,
                            t,
                            type(self).__name__
                        ))
                    else:
                        if els is None:
                            els = scaled
                        else:
                            if isinstance(scaled, (SparseArray,)):
                                scaled = scaled.asarray()
                            elif isinstance(scaled, (sp.spmatrix,)):
                                scaled = scaled.toarray()
                                # import McUtils.Plots as plt
                                #
                                # plt.ArrayPlot(scaled).show()
                            # print(scaled.shape, els.shape)
                            els += scaled
                    self.logger.log_print("took {e:.3f}s", e=(time.time() - start))
                    # this can be memory intensive so we collect each step
                    gc.collect()

        return els

    def get_brakets(self, states, check_orthogonality=True, memory_constrained=False):
        if not isinstance(states, BraKetSpace):
            states = BraKetSpace.from_indices(states, basis=self.basis)
        return self._dispatch_over_expansion('get_brakets', states, check_orthogonality=check_orthogonality, memory_constrained=memory_constrained)
    def get_element(self, n, m):
        return self._dispatch_over_expansion('get_element', n, m)
    @property
    def chunk_size(self):
        for c in self.computers:
            if c.chunk_size is not None:
                return c.chunk_size
        else:
            return None

    @property
    def selection_rules(self):
        """

        :return:
        :rtype:
        """
        if self._selection_rules is None:
            if all(hasattr(x, 'selection_rules') for x in self.computers):
                _ = []
                for x in self.computers:
                    for r in x.selection_rules:
                        if r not in _:
                            _.append(r)
                self._selection_rules = _
        return self._selection_rules

    @property
    def selection_rule_steps(self):
        """

        :return:
        :rtype:
        """
        if self._selection_rule_steps is None:
            if all(hasattr(x, 'selection_rule_steps') for x in self.computers):
                _ = []
                for x in self.computers:
                    r = x.selection_rule_steps
                    if r not in _:
                        _.append(r)
                self._selection_rule_steps = _
        return self._selection_rule_steps

    def get_transformed_space(self, space, rules=None, parallelizer=None, logger=None, **opts):
        """
        Returns the state space obtained by using the
        held operators to transform `space`

        :param space:
        :type space: BasisStateSpace
        :return:
        :rtype: SelectionRuleStateSpace
        """

        import functools

        if parallelizer is None:
            parallelizer = self.parallelizer

        # we take a union of all transformation rules and just apply that
        # if possible
        if rules is None:
            rules = self.selection_rules
        if rules is not None:
            ooooh_shiz = space.apply_selection_rules(rules, parallelizer=parallelizer, logger=logger, **opts)
        else:
            spaces = [
                r.get_transformed_space(space, parallelizer=parallelizer, logger=logger, **opts)
                for r in self.computers
            ]
            ooooh_shiz = functools.reduce(lambda s1,s2: s1.union(s2), spaces[1:], spaces[0])

        skips = Operator._get_skip_trans(self.skipped_indices, space.ndim)
        if skips is not None:
            ooooh_shiz = ooooh_shiz.filter_transitions(skips)

        return ooooh_shiz

    def __repr__(self):
        if self.name is None:
            return "{}(<{}>, ({}), ({}))".format(
                type(self).__name__,
                self._get_dim_string(self.dims) if hasattr(self, 'dims') else '???',
                self.coeffs,
                self.computers
            )
        else:
            return "{}<{}>".format(
                type(self).__name__, self.name
            )



import numpy as np

__all__ = [
    "StateMaker"
]

class StateMaker:
    """
    A tiny but useful class to make states based on their quanta
    of excitation
    """

    def __init__(self, ndim, mode='low-high'):
        self.ndim = ndim
        self.mode = mode

    @classmethod
    def parse_state(cls, state, mode='low-high'):
        nzp = np.nonzero(state)
        if len(nzp) > 0: nzp = nzp[0]
        if len(nzp) == 0: return "()"
        if mode == 'low-high':
            pos = len(state) - nzp
            state_quants = reversed(list(zip(pos, nzp)))
        elif mode == 'high-low':
            pos = nzp + 1
            state_quants = list(zip(pos, nzp))
        else:
            pos = nzp
            state_quants = list(zip(pos, nzp))
        return "".join("{}({})".format(p, state[i]) for p,i in state_quants)

    def make_state(self, *specs, mode=None):

        if mode is None:
            mode = self.mode

        state = [0] * self.ndim
        for s in specs:
            if isinstance(s, (int, np.integer)):
                i = s
                q = 1
            else:
                i,q = s
            if mode == 'low-high':
                state[-i] = q
            elif mode == 'high-low':
                state[i-1] = q
            elif mode == 'normal':
                state[i] = q
            else:
                raise ValueError("don't know what to do with filling mode '{}'".format(mode))
        return state

    def __call__(self, *specs, mode=None):
        return self.make_state(*specs, mode=mode)


"""
Defines useful extended internal coordinate frames
"""


import numpy as np
import McUtils.Numputils as nput
from McUtils.Parallelizers import Parallelizer
from McUtils.Coordinerds import (
    ZMatrixCoordinateSystem, CartesianCoordinateSystem, CoordinateSystemConverter,
    ZMatrixCoordinates, CartesianCoordinates3D, CoordinateSet, CompositeCoordinateSystem,
    GenericInternalCoordinateSystem, GenericInternalCoordinates,
    CartesianToGICSystemConverter, GICSystemToCartesianConverter
)

from ..Modes import RedundantCoordinateGenerator

from .Properties import StructuralProperties
# from .MoleculeInterface import AbstractMolecule

__all__ = [
    "MolecularEmbedding",
    "ModeEmbedding",
    "MolecularZMatrixCoordinateSystem",
    "MolecularCartesianCoordinateSystem"
]

__reload_hook__ = [".MoleculeInterface"]

class MolecularEmbedding:

    def __init__(self,
                 masses,
                 coords,
                 internals
                 ):

        self._coords = CoordinateSet(coords, MolecularCartesianCoordinateSystem(masses, coords))
        MolecularCartesianToRegularCartesianConverter(self.coords.system).register()
        RegularCartesianToMolecularCartesianConverter(self.coords.system).register()
        if isinstance(internals, CoordinateSet):
            self._int_spec = None
            self._ints = internals
        else:
            self._int_spec = self.canonicalize_internal_coordinate_spec(internals)
            self._ints = None
        self._jacobians = self._get_jacobian_storage()
        # self._int_spec = internals
        self._frame = None
        self._tr_modes = None

    @property
    def coords(self):
        return self._coords
    @coords.setter
    def coords(self, coords):
        if hasattr(coords, "system"):
            sys = coords.system
        else:
            sys = self.coords.system
            coords = CoordinateSet(coords, sys)
        if sys is not self.coords.system:
            MolecularCartesianToRegularCartesianConverter(sys).register()
            RegularCartesianToMolecularCartesianConverter(sys).register()
        self._jacobians = self._get_jacobian_storage()
        self._frame = None
        self._coords = coords
    @property
    def masses(self):
        return self._coords.system.masses

    @staticmethod
    def _wrap_conv(f):
        if f is None:
            return f

        def wrapped(*args, **kwargs):
            vals = f(*args, **kwargs)
            if not isinstance(vals, np.ndarray):
                vals, opts = vals
            else:
                opts = {}
            opts = dict(kwargs, **opts)
            return vals, opts

        return wrapped

    @classmethod
    def canonicalize_internal_coordinate_spec(cls, spec):
        if spec is not None:
            if hasattr(spec, 'items'):
                specs = spec.get('specs')
                if specs is None:
                    try:
                        zmatrix = spec['zmatrix']
                    except KeyError:
                        zmatrix = None
                    else:
                        zmatrix = np.asanyarray(zmatrix).astype(int)
                        if zmatrix.shape[1] != 4:
                            raise ValueError("can't understand Z-matrix {}".format(zmatrix))
                    spec['zmatrix'] = zmatrix
                    spec['conversion'] = cls._wrap_conv(spec.get('conversion'))
                    spec['inverse'] = cls._wrap_conv(spec.get('inverse'))
                    converter_options = spec.get('converter_options')
                    if converter_options is None:
                        converter_options = {}
                    if 'embedding_coords' not in converter_options:
                        if spec['zmatrix'] is not None:
                            converter_options['embedding_coords'] = MolecularZMatrixCoordinateSystem.embedding_coords
                    if 'jacobian_prep' not in converter_options:
                        if spec['zmatrix'] is not None:
                            converter_options['jacobian_prep'] = ZMatrixCoordinates.jacobian_prep_coordinates
                    spec['converter_options'] = converter_options
                else:
                    spec['zmatrix'] = None
                    spec['conversion'] = cls._wrap_conv(spec.get('conversion'))
                    spec['inverse'] = cls._wrap_conv(spec.get('inverse'))
            elif callable(spec):
                zmatrix = None
                specs = None
                conversion = cls._wrap_conv(spec)
                spec = {'specs': specs, 'zmatrix': zmatrix, 'conversion': conversion, 'inverse': None, 'converter_options': {}}
            else:
                if any(len(s) < 4 or hasattr(s, 'items') for s in spec):
                    conversion = None
                    specs = spec
                    zmatrix = None
                else:
                    conversion = None
                    specs = None
                    zmatrix = np.asanyarray(spec).astype(int)
                    if zmatrix.shape[1] != 4:
                        raise ValueError("can't understand Z-matrix {}".format(zmatrix))
                spec = {'specs': specs, 'zmatrix': zmatrix, 'conversion': conversion, 'inverse': None, 'converter_options': {}}
        return spec

    @property
    def internals(self):
        if self._int_spec is not None:
            return self._int_spec

    @internals.setter
    def internals(self, internals):
        self._int_spec = self.canonicalize_internal_coordinate_spec(internals)
        self._ints = None

    @property
    def zmatrix(self):
        if self._int_spec is not None:
            return self._int_spec['zmatrix']

    @zmatrix.setter
    def zmatrix(self, zmat):
        if zmat is not None:
            zmat = np.asanyarray(zmat).astype(int)
            if zmat.shape[1] != 4:
                raise ValueError("can't understand Z-matrix {}".format(zmat))
        if self._int_spec is None:
            self._int_spec = self.canonicalize_internal_coordinate_spec(zmat)
        else:
            self._int_spec['zmatrix'] = zmat
        self._ints = None

    @property
    def internal_coordinates(self):
        if self._ints is None and (
                self._int_spec is not None
                and any(self._int_spec[k] is not None for k in {'zmatrix', 'conversion', 'specs'})
        ):
            coords = self.coords
            if self._int_spec['specs'] is not None:

                opts = {
                    k:self._int_spec[k]
                    for k in [
                        'specs',
                        'redundant',
                        'untransformed_coordinates',
                        'relocalize',
                        'redundant_transformation',
                        'redundant_inverse',
                        'reference_internals',
                        'reference_coordinates'
                    ]
                    if k in self._int_spec
                }
                ints = MolecularGenericInternalCoordinateSystem(self.masses, coords, **opts
                                                                # specs=self._int_spec['specs'],
                                                                # redundant=self._int_spec.get('redundant', False),
                                                                # untransformed_coordinates=self._int_spec.get('untransformed_coordinates'),
                                                                # relocalize=self._int_spec.get('relocalize', False),
                                                                )
                MolecularCartesianToGICConverter(coords.system, ints).register()
                MolecularGICToCartesianConverter(ints, coords.system).register()
                coords = coords.convert(ints, reference_internals=opts.get('reference_internals'))
                cops = coords.converter_options
                for k in [
                    'redundant_transformation',
                    'redundant_inverse',
                    'reference_internals',
                    'reference_coordinates'
                ]:
                    if k in cops:
                        self._int_spec[k] = cops[k]
            elif self._int_spec['zmatrix'] is not None:
                zms = MolecularZMatrixCoordinateSystem(self.masses, coords, ordering=self._int_spec['zmatrix'])
                MolecularCartesianToZMatrixConverter(coords.system, zms).register()
                MolecularZMatrixToCartesianConverter(zms, coords.system).register()
                MolecularZMatrixToRegularZMatrixConverter(zms).register()
                RegularZMatrixToMolecularZMatrixConverter(zms).register()
                coords = coords.convert(zms)
            if self._int_spec['conversion'] is not None:
                conv = CompositeCoordinateSystem.register(
                    coords.system,
                    self._int_spec['conversion'],
                    inverse_conversion=self._int_spec['inverse'],
                    **self._int_spec['converter_options']
                )
                coords = coords.convert(conv)
            # print(zms)
            # print(zms, self.coords, self.coords.system.converter(zms))
            self._ints = coords
        return self._ints

    @internal_coordinates.setter
    def internal_coordinates(self, ics):
        if not isinstance(ics, CoordinateSet):
            raise ValueError("{} must be a {} to be valid internal coordinates".format(
                ics, CoordinateSet.__name__
            ))
        self._ints = ics

    def get_internals(self, strip_embedding=True):
        ics = self.internal_coordinates
        if ics is None:
            return None
        embedding_coords = self._get_embedding_coords()
        if embedding_coords is not None and strip_embedding:
            good_coords = np.setdiff1d(np.arange(3 * len(self.masses)), embedding_coords)
            ics = ics.flatten()[good_coords]
        return ics

    @property
    def redundant_internal_transformation(self):
        return self.internal_coordinates.converter_options.get('redundant_transformation')

    @classmethod
    def _get_jacobian_storage(cls):
        return {
            'internals': {"default":[], "reembed":[]},
            'fast-internals': {"default":[], "reembed":[]},
            'cartesian': []
        }
    def _get_int_jacobs(self,
                        jacs,
                        strip_dummies=False,
                        stencil=None, mesh_spacing=1.0e-3,
                        all_numerical=None, reembed=True,
                        planar_ref_tolerance=None,
                        parallelizer=None
                        ):
        """
        Gets the specified dX/dRs

        :param jacs:
        :type jacs:
        :return:
        :rtype:
        """
        intcds = self.internal_coordinates
        ccoords = self.coords
        carts = ccoords.system
        internals = intcds.system

        generics = 'GenericInternals' in internals.name
        zmatrix = 'ZMatrix' in internals.name

        if all_numerical is None:
            all_numerical = zmatrix

        converter_options = (
            dict(
                reembed=reembed,
                planar_ref_tolerance=planar_ref_tolerance,
                strip_dummies=strip_dummies
            )
                if zmatrix else
            {}
        )

        if isinstance(jacs, int):
            jacs = list(range(1, jacs + 1))

        exist_jacs = self._jacobians['internals']
        if reembed:
            exist_jacs = exist_jacs["reembed"]
        else:
            exist_jacs = exist_jacs["default"]
        # {"default": [], "reembed": [], "fast": [], "generic": []}
        max_jac = max(jacs)
        need_jacs = [x + 1 for x in range(0, max_jac) if x >= len(exist_jacs) or exist_jacs[x] is None]
        if len(need_jacs) > 0:
            if generics:
                exist_jacs[:] = [
                    x.squeeze() for x in
                    intcds.jacobian(carts, jacs,
                                    # odd behaves better
                                    mesh_spacing=mesh_spacing,
                                    stencil=stencil,
                                    all_numerical=all_numerical,
                                    converter_options=converter_options
                                    )
                ]
                # self._jacobians['internals'] = exist_jacs
            else:
                stencil = (max(need_jacs) + 2 + (1 + max(need_jacs)) % 2) if stencil is None else stencil
                # odd behaves better
                with Parallelizer.lookup(parallelizer) as par:
                    new_jacs = [
                        x.squeeze() if isinstance(x, np.ndarray) else x
                        for x in intcds.jacobian(carts, need_jacs,
                                                 # odd behaves better
                                                 mesh_spacing=mesh_spacing,
                                                 stencil=stencil,
                                                 all_numerical=all_numerical,
                                                 converter_options=converter_options,
                                                 parallelizer=par
                                                 )
                    ]
                for j, v in zip(need_jacs, new_jacs):
                    for d in range(j - len(exist_jacs)):
                        exist_jacs.append(None)
                    exist_jacs[j - 1] = v

        return [exist_jacs[j - 1] for j in jacs]

    def _get_cart_jacobs(self, jacs,
                         strip_dummies=False,
                         stencil=None, mesh_spacing=1.0e-3,
                         all_numerical=None,
                         parallelizer=None
                         ):
        """
        Gets the specified dR/dXs

        :param jacs:
        :type jacs:
        :return:
        :rtype:
        """
        intcds = self.internal_coordinates
        ccoords = self.coords
        carts = ccoords.system
        internals = intcds.system
        generics = 'GenericInternals' in internals.name
        zmatrix = 'ZMatrix' in internals.name

        if all_numerical is None:
            all_numerical = zmatrix

        converter_options = (
            dict(strip_dummies=strip_dummies)
                if zmatrix else
            {}
        )

        if isinstance(jacs, int):
            jacs = list(range(1, jacs + 1))

        exist_jacs = self._jacobians['cartesian']
        max_jac = max(jacs)
        need_jacs = [x + 1 for x in range(0, max_jac) if x >= len(exist_jacs) or exist_jacs[x] is None]
        if len(need_jacs) > 0:
            if generics:
                exist_jacs[:] = [
                    x.squeeze() for x in
                    ccoords.jacobian(internals,
                                     list(range(1, max_jac+1)),
                                     # mesh_spacing=mesh_spacing,
                                     # stencil=stencil,
                                     # all_numerical=all_numerical,
                                     converter_options=converter_options
                                     )
                ]
            else:
                stencil = (max(need_jacs) + 2 + (1 + max(need_jacs)) % 2) if stencil is None else stencil
                # odd behaves better
                with Parallelizer.lookup(parallelizer) as par:
                    exist_jacs = [
                        x.squeeze() if isinstance(x, np.ndarray) else x
                        for x in ccoords.jacobian(internals,
                                                  order=list(range(1, max_jac+1)),
                                                  mesh_spacing=mesh_spacing,
                                                  stencil=stencil,
                                                  all_numerical=all_numerical,
                                                  converter_options=converter_options,
                                                  parallelizer=par
                                                  )
                    ]

        return [exist_jacs[j - 1] for j in jacs]

    @property
    def embedding_coords(self):
        return self._get_embedding_coords()
    def _get_embedding_coords(self):
        try:
            embedding = self.internal_coordinates.system.embedding_coords
        except AttributeError:
            try:
                embedding = self.internal_coordinates.system.converter_options['embedding_coords']
            except KeyError:
                embedding = None
        return embedding

    cartesian_by_internals_method = 'fast'
    def get_cartesians_by_internals(self, order=None, strip_embedding=False, reembed=True, method=None):
        if method is None:
            int_sys = self.internal_coordinates.system
            if "GenericInternals" in int_sys.name:
                method = 'classic'
            else:
                method = self.cartesian_by_internals_method

        if reembed and method == 'fast':
            fast_ints = self._jacobians['fast-internals']["reembed"]
            if len(fast_ints) < order:
                L_base = self.get_translation_rotation_invariant_transformation(strip_embedding=strip_embedding, mass_weighted=False)
                jacs_1 = self.get_internals_by_cartesians(order, strip_embedding=strip_embedding)
                new_tf = nput.tensor_reexpand([L_base.T], jacs_1, order)
                inverse_tf = nput.inverse_transformation(new_tf, order, allow_pseudoinverse=True)
                fast_ints[:] = [
                    np.tensordot(j, L_base, axes=[-1, -1])
                    for j in inverse_tf
                ]
            return fast_ints
        elif not reembed and method == 'fast':
            # fast_ints = self._jacobians['fast-internals']["reembed"]
            wtf = self.get_internals_by_cartesians(order, strip_embedding=False) # faster to just do these derivs.
            base = nput.inverse_transformation(wtf, order-1, allow_pseudoinverse=True)
        else:
            base = (
                self._get_int_jacobs(order, reembed=reembed)
                    if order is not None else
                self._jacobians['internals']["default" if not reembed else "reembed"]
            )
            if order is not None:
                if len(base) < order:
                    raise ValueError("insufficient {} (have {} but expected {})".format(
                        'CartesiansByInternals',
                        len(base),
                        order
                    ))
                base = base[:order]

            _ = []
            sh = self.coords.shape[:-2]
            nc = 3 * len(self.masses)
            nr = len(sh)
            n = -2 if base[0].shape[-2:] == (len(self.masses), 3) else -1
            for i, b in enumerate(base):
                rem = int(np.power(np.prod(b.shape[nr:n], dtype=int), 1 / (i+1)))
                b = b.reshape(sh + (rem,) * (i+1) + (nc,))
                _.append(b)
            base = _

        embedding_coords = self._get_embedding_coords() if strip_embedding else None
        if embedding_coords is not None and strip_embedding and base[0].shape[-2] == (3 * len(self.masses)):
            good_coords = np.setdiff1d(np.arange(3 * len(self.masses)), embedding_coords)
            base = [t[np.ix_(*((good_coords,) * (t.ndim - 1)))] for t in base]
        return base

    def get_internals_by_cartesians(self, order=None, strip_embedding=False):
        base = self._get_cart_jacobs(order) if order is not None else self._jacobians['cartesian']
        if order is not None:
            if len(base) < order:
                raise ValueError("insufficient {} (have {} but expected {})".format(
                    'InternalsByCartesians',
                    len(base),
                    order
                ))
            base = base[:order]

        _ = []
        sh = self.coords.shape[:-2]
        nc = 3 * len(self.masses)
        for i, b in enumerate(base):
            b = b.reshape(sh + (nc,) * (i + 1) + (-1,))
            _.append(b)
        base = _

        if strip_embedding and base[0].shape[-1] == nc:
            embedding_coords = self._get_embedding_coords()
            if embedding_coords is not None:
                good_coords = np.setdiff1d(np.arange(3 * len(self.masses)), embedding_coords)
                base = [t[..., good_coords] for t in base]
        return base

    def embed_coords(self, coords, sel=None, in_paf=False, planar_ref_tolerance=None):
        return StructuralProperties.get_eckart_embedded_coords(
            self.masses,
            self.coords,
            coords,
            sel=sel,
            in_paf=in_paf,
            planar_ref_tolerance=planar_ref_tolerance
        )

    @property
    def inertia_tensor(self):
        return StructuralProperties.get_prop_inertia_tensors(
            self.coords,
            self.masses
        )

    @property
    def inertial_frame(self):
        """
        Provides the inertial axis frame

        :return:
        :rtype:
        """

        if self._frame is None:
            # Need to put B in Hartree?
            #  I've got moments of inertia in amu * bohr^2 at the moment
            #  So we convert (amu * bohr^2) to (m_e * bohr^2) since hb^2/(m_e bohr^2) == E_h
            mom_i, eigs = StructuralProperties.get_prop_moments_of_inertia(
                self.coords,
                self.masses
            )
            B_e = 1 / (2 * mom_i)  # * UnitsData.convert("AtomicMassUnits", "AtomicUnitOfMass"))
            self._inert_frame = B_e, eigs

        return self._inert_frame

    def inertial_frame_derivatives(self):

        return StructuralProperties.get_prop_inertial_frame_derivatives(
                self.coords,
                self.masses
            )

    @property
    def translation_rotation_modes(self):
        if self._tr_modes is None:
            self._tr_modes = StructuralProperties.get_prop_translation_rotation_eigenvectors(
                self.coords,
                self.masses
            )

        return self._tr_modes


    def get_translation_rotation_invariant_transformation(self,
                                                          order=0,
                                                          mass_weighted=True,
                                                          strip_embedding=True):
        L_tr = self.translation_rotation_modes[1]
        A = np.eye(L_tr.shape[0]) - (L_tr @ L_tr.T)
        evals, tf = np.linalg.eigh(A)
        zero_pos = np.abs(evals) < 1e-4 # the rest should be 1
        tf[:, zero_pos] = L_tr
        if strip_embedding:
            nzpos = np.abs(evals) > 1e-4
            tf = tf[:, nzpos]

        if not mass_weighted:
            tf = np.diag(np.repeat(1/np.sqrt(self.masses), 3)) @ tf

        return tf

class ModeEmbedding:
    """
    Provides a specialization on a `MoleculaEmbedding` to express all properties
    in terms of the attendant normal modes
    """
    def __init__(self,
                 embedding: MolecularEmbedding,
                 modes,  #:#NormalModesManager,
                 mass_weight=None,
                 dimensionless=None,
                 masses=None
                 ):
        self.embedding = embedding
        self.masses = masses
        if hasattr(modes, 'modes'):
            modes = modes.modes
        if hasattr(modes, 'basis') and hasattr(modes.basis, 'to_new_modes'):
            modes = modes.basis.to_new_modes()
        if modes is not None:
            if dimensionless:
                modes = modes.make_dimensionless(masses=masses)
            elif mass_weight:
                modes = modes.make_mass_weighted(masses=masses)

        if modes is not None:
            self.mass_weighted = modes.mass_weighted
        else:
            self.mass_weighted = True
        self.modes = modes

    def mw_conversion(self, strip_dummies=None):
        masses = self.masses
        if masses is None:
            masses = self.embedding.masses
        if strip_dummies:
            masses = masses[masses > 0]
        mvec = np.broadcast_to(
                np.asanyarray(masses)[:, np.newaxis],
                (len(masses), 3)
            ).flatten()
        return np.diag(np.sign(mvec) * np.sqrt(np.abs(mvec)))
    def mw_inverse(self, strip_dummies=None):
        masses = self.masses
        if masses is None:
            masses = self.embedding.masses
        if strip_dummies:
            masses = masses[masses > 0]
        mvec = np.broadcast_to(
                np.asanyarray(masses)[:, np.newaxis],
                (len(masses), 3)
            ).flatten()
        return np.diag(np.sign(mvec) / np.sqrt(np.abs(mvec)))

    def get_mw_cartesians_by_internals(self, order=None, strip_embedding=True):
        RX = self.embedding.get_cartesians_by_internals(
                order=order,
                strip_embedding=strip_embedding
            )
        if self.mass_weighted:
            XY = self.mw_conversion()
            RX = [
                np.tensordot(tf_X, XY, axes=[-1, 0])
                for tf_X in RX
            ]
        return RX
    def get_internals_by_mw_cartesians(self, order=None, strip_embedding=True):
        XR = self.embedding.get_internals_by_cartesians(
                order=order,
                strip_embedding=strip_embedding
            )
        if self.mass_weighted:
            YX = self.mw_inverse()
            YR = []
            for X_tf in XR:
                for d in range(X_tf.ndim-1):
                    X_tf = np.tensordot(YX, X_tf, axes=[1, d])
                YR.append(X_tf)
        else:
            YR = XR
        return YR

    def get_internals_by_cartesians(self, order=None, strip_embedding=True):
        """
        expresses raw internals or modes (internals or Cartesian) in terms of mass-weighted Cartesians

        :param order:
        :param strip_embedding:
        :return:
        """
        if self.embedding.internals is None:
            if self.modes is None:
                raise NotImplementedError("not sure what's most consistent for just...plain Cartesians")
            return [self.modes.modes_by_coords]
        else:
            if self.modes is None or self.modes.is_cartesian:
                if self.mass_weighted:
                    YR = self.get_internals_by_mw_cartesians(
                        order=order,
                        strip_embedding=strip_embedding
                    )
                else:
                    YR = self.embedding.get_internals_by_cartesians(
                        order=order,
                        strip_embedding=strip_embedding
                    )
                if self.modes is not None:
                    YQ = self.modes.modes_by_coords
                    if self.mass_weighted:
                        RY = self.get_mw_cartesians_by_internals(
                            order=1,
                            strip_embedding=strip_embedding
                        )[0]
                    else:
                        RY = self.embedding.get_cartesians_by_internals(
                            order=1,
                            strip_embedding=strip_embedding
                        )[0]
                    RQ = RY @ YQ
                    YR = nput.tensor_reexpand(YR, [RQ], order=order)
                return YR
            else:
                tens = self.embedding.get_internals_by_cartesians(order)
                return nput.tensor_reexpand(
                    tens,
                    [self.modes.modes_by_coords],
                    order=order if order is not None else len(tens)
                )

    def get_cartesians_by_internals(self, order=None, strip_embedding=True):
        """
        expresses raw internals or modes (internals or Cartesian) in terms of mass-weighted Cartesians

        :param order:
        :param strip_embedding:
        :return:
        """
        if self.embedding.internals is None:
            if self.modes is None:
                raise NotImplementedError("not sure what's most consistent for just...plain Cartesians")
            return [self.modes.coords_by_modes]
        else:
            if self.modes is None or self.modes.is_cartesian:
                if self.modes is not None: strip_embedding = True
                if self.mass_weighted:
                    RY = self.get_mw_cartesians_by_internals(
                        order=order,
                        strip_embedding=strip_embedding
                    )
                else:
                    RY = self.embedding.get_cartesians_by_internals(
                        order=order,
                        strip_embedding=strip_embedding
                    )
                if self.modes is not None:
                    QY = self.modes.coords_by_modes
                    if self.mass_weighted:
                        YR = self.get_internals_by_mw_cartesians(
                            order=1,
                            strip_embedding=strip_embedding
                        )[0]
                    else:
                        YR = self.embedding.get_internals_by_cartesians(
                            order=1,
                            strip_embedding=strip_embedding
                        )[0]
                    QR = QY @ YR
                    RY = nput.tensor_reexpand([QR], RY, order=order)
                return RY
            else:
                tens = self.embedding.get_cartesians_by_internals(order),
                return nput.tensor_reexpand(
                    [self.modes.coords_by_modes],
                    tens,
                    order=order if order is not None else len(tens)
                )

    def get_inertia_tensor_expansion(self, order=None, strip_embedding=True):
        YI0 = self.embedding.inertial_frame_derivatives()
        QY = self.get_cartesians_by_internals(order=order, strip_embedding=strip_embedding)
        return [self.embedding.inertia_tensor] + nput.tensor_reexpand(QY, YI0, order=order)

    def get_inertial_frame(self):
        return self.embedding.inertial_frame

def _get_best_axes(first_pos, axes):
    """
    Determine the best pair of inertial axes so that we don't get large-scale breakdowns from the choice of embedding

    :param first_pos:
    :type first_pos:
    :param axes:
    :type axes:
    :return:
    :rtype:
    """

    if axes.ndim > 2:
        axes = axes[..., (0, 1), :]
        ax_choice = (0, 1)
        ax_names = ["A", "B"]
    else:
        fp_norm = np.linalg.norm(first_pos)
        if fp_norm > 1.0e-10:  # not chilling at the origin...
            first_pos = first_pos / fp_norm
            # check if it lies along an axis or is perpendicular to an axis
            a_proj = np.dot(first_pos, axes[0])
            b_proj = np.dot(first_pos, axes[1])
            c_proj = np.dot(first_pos, axes[2])
            if np.abs(b_proj) < .05: # lies in the A/C plane
                if np.abs(a_proj) > .95:
                    ax_choice = (1, 2)
                    ax_names = ["B", "C"]
                else:
                    ax_choice = (0, 1)
                    ax_names = ["A", "B"]
            elif np.abs(c_proj) < .05: # lies in the A/B plane
                if np.abs(a_proj) > .95:
                    ax_choice = (1, 2)
                    ax_names = ["B", "C"]
                else:
                    ax_choice = (0, 2)
                    ax_names = ["A", "C"]
            elif np.abs(a_proj) < .05:  # lies in the B/C plane
                if np.abs(b_proj) > .95:
                    ax_choice = (0, 2)
                    ax_names = ["A", "C"]
                else:
                    ax_choice = (0, 1)
                    ax_names = ["A", "B"]
            else: # not in any of the planes so no issues
                ax_choice = (0, 1)
                ax_names = ["A", "B"]

        else:
            ax_choice = (0, 1)
            ax_names = ["A", "B"]
        axes = axes[ax_choice,]
    return axes, ax_names, ax_choice

class MolecularGenericInternalCoordinateSystem(GenericInternalCoordinateSystem):
    """
    Mirrors the standard ZMatrix coordinate system in _almost_ all regards, but forces an embedding
    """
    name = "MolecularGenericInternals"
    class PassThroughRedundantGenerator:
        def __init__(self, redundant_transformation, redundant_inverse=None,
                     masses=None,
                     untransformed_coordinates=None,
                     relocalize=False,
                     **opts
                     ):
            self.tf = redundant_transformation
            self.inv = redundant_inverse
            self.masses = masses,
            self.untransformed_coordinates = untransformed_coordinates
            self.relocalize = relocalize
            self.opts = opts
        def get_redundant_transformation(self, base_expansions, **ignored):
            return self.tf, nput.tensor_reexpand(base_expansions,
                                                 [self.tf],
                                                 order=len(base_expansions)
                                                 )
    def __init__(self, masses, coords, /, specs, converter_options=None,
                 redundant=False,
                 relocalize=False,
                 untransformed_coordinates=None,
                 redundant_transformation=None,
                 redundant_inverse=None,
                 angle_ordering='ijk',
                 **opts):
        """

        :param molecule:
        :type molecule: AbstractMolecule
        :param converter_options:
        :type converter_options:
        :param opts:
        :type opts:
        """
        self.masses = masses
        self.coords = coords
        if converter_options is None:
            converter_options = dict(opts, reference_coordinates=coords, masses=masses, angle_ordering=angle_ordering, specs=specs)
            opts = {}
        if redundant and 'redundant_generator' not in converter_options:
            if redundant_transformation is not None:
                redundant_generator = self.PassThroughRedundantGenerator(
                    redundant_transformation,
                    redundant_inverse=redundant_inverse,
                    masses=self.masses,
                    untransformed_coordinates=untransformed_coordinates,
                    angle_ordering=angle_ordering,
                    relocalize=relocalize
                )
            else:
                redundant_generator = RedundantCoordinateGenerator(
                    specs,
                    masses=self.masses,
                    untransformed_coordinates=untransformed_coordinates,
                    angle_ordering=angle_ordering,
                    relocalize=relocalize
                )
            converter_options['redundant_generator'] = converter_options.get('redundant_generator', redundant_generator)
        nats = None#len(specs)
        super().__init__(converter_options=converter_options, dimension=(nats,), coordinate_shape=(nats,), opts=opts)

class MolecularZMatrixCoordinateSystem(ZMatrixCoordinateSystem):
    """
    Mirrors the standard ZMatrix coordinate system in _almost_ all regards, but forces an embedding
    """
    name = "MolecularZMatrix"
    embedding_coords = [0, 1, 2, 4, 5, 8]
    def __init__(self, masses, coords, converter_options=None, **opts):
        """

        :param molecule:
        :type molecule: AbstractMolecule
        :param converter_options:
        :type converter_options:
        :param opts:
        :type opts:
        """
        self.masses = masses
        self.coords = coords
        self.dummy_positions = [i for i, m in enumerate(masses) if m < 0]
        self.com = StructuralProperties.get_prop_center_of_mass(coords, masses)
        self.inertial_axes = StructuralProperties.get_prop_moments_of_inertia(coords, masses)[1]
        if converter_options is None:
            converter_options = opts
            opts = {}
        nats = len(masses)
        super().__init__(converter_options=converter_options, dimension=(nats, 3), coordinate_shape=(nats, 3), opts=opts)
        self.set_embedding()
    @property
    def origins(self):
        return self.converter_options['origins']
    @property
    def axes(self):
        return self.converter_options['axes']

    def pre_convert(self, system):
        # self.converter_options['molecule'] = self.molecule
        self.set_embedding()

    def set_embedding(self):
        com = self.com
        axes = self.inertial_axes
        converter_options = self.converter_options
        if 'ordering' in converter_options:
            ordering = np.array(converter_options['ordering'], dtype=int)
            ordering[0, 1] = -3; ordering[0, 2] = -1; ordering[0, 3] = -2
            ordering[1, 2] = -1; ordering[1, 3] = -2
            if len(ordering) > 2:
                ordering[2, 3] = -2
            converter_options['ordering'] = ordering
            first = ordering[0, 0]
        else:
            first = 0

        first_pos = self.coords[first]
        axes, ax_names, ax_choice = _get_best_axes(first_pos, axes)

        converter_options['origins'] = com
        converter_options['axes'] = axes
        converter_options['axes_labels'] = ax_names
        converter_options['axes_choice'] = ax_choice
        converter_options['masses'] = self.masses
        converter_options['dummy_positions'] = self.dummy_positions
        converter_options['ref_coords'] = self.coords

    def jacobian(self,
                 coords,
                 *args,
                 reembed=None,
                 strip_dummies=None,
                 converter_options=None,
                 **kwargs
                 ):

        if converter_options is None:
            converter_options = {}
        merged_convert_options = dict(self.converter_options, **converter_options)
        try:
            remb = merged_convert_options['reembed'] if reembed is None else reembed
        except KeyError:
            remb = None

        try:
            strip_dummies = merged_convert_options['strip_dummies'] if strip_dummies is None else strip_dummies
        except KeyError:
            strip_dummies = False

        if strip_dummies:
            dummies = self.dummy_positions
        else:
            dummies = None

        if dummies is not None:
            main_excludes = np.setdiff1d(
                np.arange(self.masses),
                dummies
            )
        else:
            main_excludes = None

        if coords.ndim > 3:
            og_shape = coords.shape
            coords = coords.reshape((-1,) + coords.shape[-2:])
        else:
            og_shape = (coords.shape[0],) if coords.ndim == 3 else ()

        # print("...?", coords.shape)

        try:
            self.converter_options['reembed'] = True if remb is None else remb
            jacs = super().jacobian(coords, *args, converter_options=converter_options, **kwargs)
            if isinstance(jacs, np.ndarray):
                jacs = [jacs]
            raw_jacs = []
            for j in jacs:
                skip_dim = coords.ndim - 2
                if skip_dim > 0:
                    j = np.moveaxis(j, -3, 0) # skip_dim == 1 by construction so only need 1 move...
                ext_dim = j.ndim - 2 - skip_dim
                # print("????", j.shape, coords.shape, skip_dim)
                shp = og_shape + sum(
                    ((j.shape[i] // 3, 3) for i in range(skip_dim, ext_dim+skip_dim)),
                    ()
                ) + j.shape[-2:]
                j = j.reshape(shp)
                if main_excludes is not None:
                    for i in range(ext_dim):
                        j = np.take(j, main_excludes, axis=2*i)
                raw_jacs.append(j)
            jacs = raw_jacs
            return jacs
        finally:
            if remb is not None:
                self.converter_options['reembed'] = remb

class MolecularCartesianCoordinateSystem(CartesianCoordinateSystem):
    """
    Mirrors the standard Cartesian coordinate system in _almost_ all regards, but forces an embedding
    """
    name= "MolecularCartesians"
    def __init__(self, masses, coords, dummy_positions=None, converter_options=None, **opts):
        """

        :param molecule:
        :type molecule: AbstractMolecule
        :param converter_options:
        :type converter_options:
        :param opts:
        :type opts:
        """
        self.masses = masses
        self.coords = coords
        self.dummy_positions = [i for i,m in enumerate(masses) if m < 0] if dummy_positions is None else dummy_positions
        self.com = StructuralProperties.get_prop_center_of_mass(coords, masses)
        self.axes = StructuralProperties.get_prop_moments_of_inertia(coords, masses)[1]
        nats = len(masses)
        if converter_options is None:
            converter_options = opts
            opts = {}
        super().__init__(converter_options=converter_options, dimension=(nats, 3), coordinate_shape=(nats, 3), opts=opts)

    def pre_convert(self, system):
        self.converter_options['masses'] = self.masses
        if 'ZMatrix' in system.name:
            self.set_embedding()

    def set_embedding(self):
        """
        Sets up the embedding options...
        :return:
        :rtype:
        """
        com = self.com
        axes = self.axes
        converter_options = self.converter_options
        if 'ordering' in converter_options:
            ordering = np.array(converter_options['ordering'], dtype=int)
            ordering[0, 1] = -3; ordering[0, 2] = -2; ordering[0, 3] = -1
            ordering[1, 2] = -1; ordering[1, 3] = -2
            if len(ordering) > 2:
                ordering[2, 3] = -2
            converter_options['ordering'] = ordering
            first = ordering[0, 0]
        else:
            first = 0

        first_pos = self.coords[first]
        axes, ax_names, ax_choice = _get_best_axes(first_pos, axes)

        converter_options['origins'] = com
        converter_options['axes'] = axes
        converter_options['axes_labels'] = ax_names
        converter_options['axes_choice'] = ax_choice
        converter_options['masses'] = self.masses
        converter_options['ref_coords'] = self.coords
        converter_options['dummy_positions'] = self.dummy_positions

    def jacobian(self,
                 coords,
                 system,
                 order=None,
                 strip_dummies=None,
                 converter_options=None,
                 analytic_deriv_order=None,
                 **kwargs
                 ):

        zmat_conv = 'ZMatrix' in system.name

        if converter_options is None:
            converter_options = {}
        merged_convert_options = dict(self.converter_options, **converter_options)
        try:
            strip_dummies = merged_convert_options['strip_dummies'] if strip_dummies is None else strip_dummies
        except KeyError:
            strip_dummies = False

        try:
            analytic_deriv_order = merged_convert_options['analytic_deriv_order'] if analytic_deriv_order is None else analytic_deriv_order
        except KeyError:
            if zmat_conv:
                analytic_deriv_order = 0
            else:
                # if return_derivs is None:
                #     return_derivs = order
                analytic_deriv_order = None #return_derivs if nput.is_numeric(return_derivs) else max(return_derivs)

        if strip_dummies:
            dummies = self.dummy_positions
            if len(dummies) == 0:
                dummies = None
        else:
            dummies = None

        if dummies is not None:
            main_excludes = np.setdiff1d(
                np.arange(len(self.masses)),
                dummies
            )
        else:
            main_excludes = None

        jacs = super().jacobian(coords, system,
                                order=order,
                                analytic_deriv_order=analytic_deriv_order,
                                converter_options=converter_options,
                                **kwargs)
        if isinstance(jacs, np.ndarray):
            jacs = [jacs]

        if zmat_conv:
            raw_jacs = []
            for n,j in enumerate(jacs): # this expects a full filling of the jacobians which maybe I need to not expect...
                baseline = 2*analytic_deriv_order + len(coords.shape)
                ext_dim = j.ndim - baseline
                shp = sum(
                    ((j.shape[i] // 3, 3) for i in range(ext_dim)),
                    ()
                ) + j.shape[-baseline:]
                j = j.reshape(shp)
                if dummies is not None:
                    for i in range(ext_dim):
                        j = np.take(j, main_excludes, axis=2*i)
                    for i in range(analytic_deriv_order):
                        j = np.take(j, main_excludes, axis=-2*(i+2))

                if len(coords.shape) > 2:
                    j = np.moveaxis(j, -3, 0)

                raw_jacs.append(j)
            jacs = raw_jacs
        return jacs

class MolecularCartesianToZMatrixConverter(CoordinateSystemConverter):
    """
    ...
    """
    def __init__(self, cart_system, zmat_system, **opts):
        self._types = (cart_system, zmat_system)
        super().__init__(**opts)
    @property
    def types(self):
        return self._types
    def convert(self, coords, *,
                     masses,
                     dummy_positions, origins=None, axes=None, ordering=None, **kwargs):
        """
        Converts from Cartesian to ZMatrix coords, preserving the embedding
        :param coords:
        :type coords: CoordinateSet
        :param molecule:
        :type molecule:
        :param origins:
        :type origins:
        :param axes:
        :type axes:
        :param ordering:
        :type ordering:
        :param kwargs:
        :type kwargs:
        :return:
        :rtype:
        """

        zmcs, opts = self.convert_many(np.array([coords]),
                                       masses=masses,
                                       dummy_positions=dummy_positions,
                                       origins=origins, axes=axes, ordering=ordering, **kwargs)
        zmcs = zmcs[0]

        if 'derivs' in opts:
            derivs = opts['derivs']
            reshaped_derivs = [None] * len(derivs)
            for i, v in enumerate(derivs):
                reshaped_derivs[i] = v[0]
            opts['derivs'] = reshaped_derivs

        return zmcs, opts

    def convert_many(self,
                     coords,*,
                     masses,
                     dummy_positions,
                     origins=None, axes=None,
                     ordering=None,
                     strip_embedding=True,
                     strip_dummies=False,
                     return_derivs=None,
                     **kwargs):
        """
        Converts from Cartesian to ZMatrix coords, preserving the embedding

        :param coords: coordinates in Cartesians to convert
        :type coords: np.ndarray
        :param molecule:
        :type molecule: MolecularEmbedding
        :param origins: the origin for each individual structure
        :type origins: np.ndarray
        :param axes: the axes for each structure
        :type axes: np.ndarray
        :param ordering: the Z-matrix ordering spec
        :type ordering:
        :param strip_embedding: whether to strip the embedding coordinates
        :type strip_embedding:
        :param strip_dummies: whether to strip all dummy coordinates
        :type strip_dummies:
        :param kwargs:
        :type kwargs:
        :return:
        :rtype:
        """
        return_derivs=False

        n_sys = coords.shape[0]
        n_coords = coords.shape[1]
        n_atoms = len(masses)

        # we add three dummy atoms at the origins and along the axes before doing the conversion
        if origins.ndim == 1:
            origins = np.broadcast_to(origins[np.newaxis, np.newaxis], (n_sys, 1, 3))
        elif origins.ndim == 2:
            origins = origins[:, np.newaxis, :]
        if axes.ndim == 2:
            axes = np.broadcast_to(axes[np.newaxis], (n_sys, 2, 3))
        if origins.shape[0] != n_sys:
            if n_sys % origins.shape[0] != 0:
                raise ValueError("inconsistent shapes; origins shape {} but coords shape {}".format(
                    origins.shape,
                    coords.shape
                ))
            num_coords = n_sys // origins.shape[0]
            origins = np.broadcast_to(origins[:, np.newaxis, :, :], (origins.shape[0], num_coords) + origins.shape[1:])
            origins = origins.reshape((n_sys,) + origins.shape[2:])
        if axes.shape[0] != n_sys:
            if n_sys % axes.shape[0] != 0:
                raise ValueError("inconsistent shapes; axes shape {} but coords shape {}".format(
                    axes.shape,
                    coords.shape
                ))
            num_coords = n_sys // axes.shape[0]
            axes = np.broadcast_to(axes[:, np.newaxis, :, :], (axes.shape[0], num_coords) + axes.shape[1:])
            axes = axes.reshape((n_sys,) + axes.shape[2:])
        coords = np.concatenate([origins, origins+axes, coords], axis=1)
        if ordering is not None:
            ordering = np.array(ordering, dtype=int)
            ordering[0, 1] = -3; ordering[0, 2] = -2; ordering[0, 3] = -1
            ordering[1, 2] = -2; ordering[1, 3] = -1
            if len(ordering) > 2:
                ordering[2, 3] = -1
            ordering = ordering + 3
            ordering = np.concatenate([ [[0, -1, -1, -1], [1, 0, -1, -1], [2, 0, 1, -1]], ordering])
        res = CoordinateSet(coords, CartesianCoordinates3D).convert(ZMatrixCoordinates,
                                                                    ordering=ordering,
                                                                    origins=origins,
                                                                    axes=axes,
                                                                    return_derivs=return_derivs,
                                                                    **kwargs
                                                                    )

        if isinstance(res, tuple):
            zmcs, opts = res
        else:
            zmcs = res
            opts=res.converter_options
        opts['ordering'] = opts['ordering'][3:] - 3
        # zmcs = zmcs[:, 2:]
        if strip_dummies:
            dummies = [0, 1, 2] + [x+3 for x in dummy_positions] # add on axes
        elif strip_embedding:
            dummies = [0, 1, 2]
        else:
            dummies = None

        if dummies is not None:
            main_excludes = np.setdiff1d(
                np.arange(n_atoms + 3),
                dummies
            )
            sub_excludes = main_excludes - 1 # drop one fewer terms to drop I think...
            if 'derivs' in opts:
                derivs = opts['derivs']
                reshaped_derivs = [None] * len(derivs)
                deriv_excludes = np.arange(3, n_atoms + 3)
                for i, v in enumerate(derivs):
                    # drop all terms relating to the embedding of the embedding
                    start_dim = v.ndim - 2*(i+2)
                    for j in range(start_dim, v.ndim-2, 2):
                        v = np.take(v, deriv_excludes, axis=j)
                    v = np.take(v, sub_excludes, axis=-2)
                    reshaped_derivs[i] = v

                opts['derivs'] = reshaped_derivs

            zmcs = zmcs[..., sub_excludes, :]
        return zmcs, opts
# MolecularCartesianToZMatrixConverter = MolecularCartesianToZMatrixConverter()
# MolecularCartesianToZMatrixConverter.register(CoordinateSystemConverters)
class MolecularCartesianToRegularCartesianConverter(CoordinateSystemConverter):
    """
    ...
    """

    def __init__(self, cart_system, **opts):
        self._types = (CartesianCoordinates3D, cart_system)
        super().__init__(**opts)
    @property
    def types(self):
        return self._types
    def convert(self, coords, **kw):
        return coords, kw

    def convert_many(self, coords, **kwargs):
        """
        Converts from Cartesian to ZMatrix coords, preserving the embedding
        """
        return coords, kwargs
# MolecularCartesianToRegularCartesianConverter = MolecularCartesianToRegularCartesianConverter()
# MolecularCartesianToRegularCartesianConverter.register()
class RegularCartesianToMolecularCartesianConverter(CoordinateSystemConverter):
    """
    ...
    """

    def __init__(self, cart_system, **opts):
        self._types = (cart_system, CartesianCoordinates3D)
        super().__init__(**opts)
    @property
    def types(self):
        return self._types

    def convert(self, coords, **kw):
        return coords, kw

    def convert_many(self, coords, **kwargs):
        """
        Converts from Cartesian to ZMatrix coords, preserving the embedding
        """
        return coords, kwargs

class MolecularZMatrixToCartesianConverter(CoordinateSystemConverter):
    """
    ...
    """
    def __init__(self, zmat_system, cart_system, **opts):
        self._types = (zmat_system, cart_system)
        super().__init__(**opts)
    @property
    def types(self):
        return self._types

    def convert(self, coords, **kw):
        total_points, opts = self.convert_many(coords[np.newaxis], **kw)
        return total_points[0], opts

    def convert_many(self, coords, *,
                     masses, dummy_positions, ref_coords,
                     origins=None, axes=None, ordering=None,
                     reembed=False, axes_choice=None, return_derivs=None,
                     strip_dummies=False,
                     strip_embedding=True,
                     planar_ref_tolerance=None,
                     **kwargs):
        """
        Converts from Cartesian to ZMatrix coords, attempting to preserve the embedding
        """
        from .Molecule import Molecule
        return_derivs = False

        n_sys = coords.shape[0]
        n_coords = coords.shape[1]
        n_atoms = len(masses)
        if n_coords != n_atoms + 2:
            # means we already added the embedding
            if n_coords != n_atoms:
                raise ValueError('Embedding unclear when num_coords ({}) < num_atoms ({})'.format(
                    n_coords,
                    n_atoms
                ))

            # TODO:
            # these need broadcasting when
            # doing jacobian calcs but this is currently pretty
            # hacky
            # print(axes.shape, n_sys)
            if axes.ndim == 2:
                axes = axes[np.newaxis]
            if axes.shape[0] > 1 and axes.shape[0] < n_sys:
                # print((n_sys // axes.shape[0], axes.shape[0]) + axes.shape)
                axes = np.reshape(
                    np.broadcast_to(
                        axes[np.newaxis],
                        (n_sys // axes.shape[0],) + axes.shape
                    ),
                    (n_sys, ) + axes.shape[1:]
                )

            if origins.ndim == 1:
                origins = origins[np.newaxis]
            # print(axes.shape, n_sys)
            if origins.shape[0] > 1 and origins.shape[0] < n_sys:
                # print((n_sys // axes.shape[0], axes.shape[0]) + axes.shape)
                origins = np.reshape(
                    np.broadcast_to(
                        origins[np.newaxis],
                        (n_sys // origins.shape[0],) + origins.shape
                    ),
                    (n_sys, ) + origins.shape[1:]
                )

            x_ax = axes[..., 0, :]
            y_ax = axes[..., 1, :]
            extra_norms0 = nput.vec_norms(x_ax)
            extra_norms1 = nput.vec_norms(y_ax)
            extra_angles, _ = nput.vec_angles(x_ax, y_ax)

            extra_coords = np.zeros((n_sys, 2, 3))
            extra_coords[..., 0, 0] = extra_norms0
            extra_coords[..., 1, 0] = extra_norms1
            extra_coords[..., 1, 1] = extra_angles

            coords = np.concatenate([extra_coords, coords], axis=-2)
            if ordering is not None:
                ordering = np.array(ordering, dtype=int)
                ordering = ordering + 3
                ordering = np.concatenate([ [[0, -1, -1, -1], [1, 0, -1, -1], [2, 0, 1, -1]], ordering])

        refuse_derivs = reembed and coords.squeeze().ndim != 2
        res = CoordinateSet(coords, ZMatrixCoordinates).convert(CartesianCoordinates3D,
                                                                        ordering=ordering,
                                                                        origins=origins,
                                                                        axes=axes,
                                                                        return_derivs=(return_derivs and not refuse_derivs),
                                                                        **kwargs)

        if isinstance(res, tuple):
            carts, opts = res
        else:
            carts = res
            opts = res.converter_options

        if reembed:
            embed_carts = carts[..., 3:, :]
            reembed = not (
                    carts.squeeze().ndim == 2 and
                    np.allclose(ref_coords, embed_carts, atol=1.0e-5)
            ) # agree to like a ten thousandth of an angstrom
            if reembed:
                if not return_derivs:
                    embed_carts = StructuralProperties.get_eckart_embedded_coords(
                        masses,
                        ref_coords,
                        embed_carts,
                        planar_ref_tolerance=planar_ref_tolerance
                    )
                    carts = np.concatenate([
                        carts[..., :3, :],
                        embed_carts
                        ],
                    axis=-2
                    )
                else:
                    inert_coords, coord_coms, coord_axes = StructuralProperties.get_prop_principle_axis_data(
                        coords,
                        masses
                    ).principle_axis_data
                    if axes_choice is None:
                        axes_choice = (0, 1)
                    guh = self.convert_many(coords,
                                            origins=coord_coms,
                                            axes=coord_axes[:, axes_choice],
                                            masses=masses,
                                            ref_coords=ref_coords,
                                            dummy_positions=dummy_positions,
                                            reembed=False,
                                            ordering=ordering,
                                            return_derivs=return_derivs,
                                            axes_choice=axes_choice,
                                            **kwargs
                                            )
                    return guh

        opts['origins'] = origins
        opts['axes'] = axes
        if ordering is not None:
            opts['ordering'] = ordering[3:] - 3
        if strip_dummies:
            dummies = [0, 1, 2] + [x + 3 for x in dummy_positions]  # add on axes
        elif strip_embedding:
            dummies = [0, 1, 2]
        else:
            dummies = None
        if dummies is not None:
            main_excludes = np.setdiff1d(
                np.arange(len(masses) + 3),
                dummies
            )
            sub_excludes = main_excludes - 1  # drop one fewer terms to drop I think...
            if 'derivs' in opts:
                derivs = opts['derivs']
                reshaped_derivs = [None] * len(derivs)
                deriv_excludes = np.arange(3, len(masses) + 3)
                for i, v in enumerate(derivs):
                    # drop all terms relating to the embedding of the embedding
                    start_dim = v.ndim - i
                    for j in range(start_dim, v.ndim, 2):
                        v = np.take(v, deriv_excludes, axis=j)
                    v = np.take(v, sub_excludes, axis=-2)
                    reshaped_derivs[i] = v
                opts['derivs'] = reshaped_derivs

            carts = carts[..., main_excludes, :]

        return carts, opts
# MolecularZMatrixToCartesianConverter = MolecularZMatrixToCartesianConverter()
# MolecularZMatrixToCartesianConverter.register()

class MolecularZMatrixToRegularZMatrixConverter(CoordinateSystemConverter):
    """
    ...
    """
    def __init__(self, zmat_system, **opts):
        self._types = (zmat_system, ZMatrixCoordinateSystem)
        super().__init__(**opts)
    @property
    def types(self):
        return self._types

    def convert(self, coords, **kw):
        return coords, kw

    def convert_many(self, coords, **kwargs):
        return coords, kwargs

class RegularZMatrixToMolecularZMatrixConverter(CoordinateSystemConverter):
    """
    ...
    """

    def __init__(self, zmat_system, **opts):
        self._types = (ZMatrixCoordinateSystem, zmat_system)
        super().__init__(**opts)
    @property
    def types(self):
        return self._types

    def convert(self, coords, **kw):
        return coords, kw

    def convert_many(self, coords, **kwargs):
        return coords, kwargs
# MolecularZMatrixToRegularZMatrixConverter = MolecularZMatrixToRegularZMatrixConverter()
# MolecularZMatrixToRegularZMatrixConverter.register()

class MolecularCartesianToGICConverter(CartesianToGICSystemConverter):
    """
    ...
    """
    def __init__(self, cart_system, zmat_system, **opts):
        self._types = (cart_system, zmat_system)
        super().__init__(**opts)
    @property
    def types(self):
        return self._types

    def convert_many(self, coords, *,
                     order=0,
                     return_derivs=None,
                     redundant_generator:RedundantCoordinateGenerator=None,
                     reference_internals=None,
                     **kw):
        """
        We'll implement this by having the ordering arg wrap around in coords?
        """
        internals, opts = super().convert_many(coords, order=order, return_derivs=return_derivs, **kw)
        if redundant_generator is not None:
            red_tf, red_exp = redundant_generator.get_redundant_transformation(
                opts['derivs'],
                untransformed_coordinates=redundant_generator.untransformed_coordinates,
                masses=redundant_generator.masses,
                relocalize=redundant_generator.relocalize
            )
            if reference_internals is None:
                opts['reference_internals'] = internals
                internals = np.zeros(internals.shape[:-1] + red_tf.shape[-1:], dtype=internals.dtype)
            else:
                internals = nput.vec_tensordot(
                    red_tf,
                    internals - reference_internals,
                    axes=[-2, 0]
                )
            # internals = internals @ red_tf
            opts['derivs'] = red_exp
            opts['redundant_transformation'] = red_tf
            opts['redundant_inverse'] = np.moveaxis(red_tf, -1, -2)

        return internals, opts
# MolecularCartesianToZMatrixConverter = MolecularCartesianToZMatrixConverter()
# MolecularCartesianToZMatrixConverter.register(CoordinateSystemConverters)

class MolecularGICToCartesianConverter(GICSystemToCartesianConverter):
    """
    ...
    """
    def __init__(self, cart_system, zmat_system, **opts):
        self._types = (cart_system, zmat_system)
        super().__init__(**opts)
    @property
    def types(self):
        return self._types
    def convert_many(self,
                     coords, *,
                     redundant_transformation=None,
                     redundant_inverse=None,
                     redundant_generator=None,
                     reference_internals=None,
                     **kw):
        # if redundant_inverse is None and redundant_transformation is not None:
        #     redundant_inverse = np.moveaxis(redundant_transformation, -1, -2)
        carts, opts = super().convert_many(coords,
                                           base_transformation=(
                                               [redundant_transformation]
                                               if redundant_transformation is not None else
                                               None
                                           ),
                                           reference_internals=reference_internals,
                                           **kw)
        # if redundant_transformation is not None and 'derivs' in opts:
        #     opts['derivs'] = nput.tensor_reexpand([redundant_inverse], opts['derivs'])

        return carts, opts

class RegularGICToMolecularGICConverter(CoordinateSystemConverter):
    """
    ...
    """

    def __init__(self, **opts):
        self._types = (GenericInternalCoordinates, MolecularGenericInternalCoordinateSystem)
        super().__init__(**opts)
    @property
    def types(self):
        return self._types

    def convert(self, coords, **kw):
        return coords, kw

    def convert_many(self, coords, **kwargs):
        return coords, kwargs
# MolecularZMatrixToRegularZMatrixConverter = MolecularZMatrixToRegularZMatrixConverter()
# MolecularZMatrixToRegularZMatrixConverter.register()

class MolecularGICConverterToRegularGIC(CoordinateSystemConverter):
    """
    ...
    """

    def __init__(self, **opts):
        self._types = (MolecularGenericInternalCoordinateSystem, GenericInternalCoordinates)
        super().__init__(**opts)
    @property
    def types(self):
        return self._types

    def convert(self, coords, **kw):
        return coords, kw

    def convert_many(self, coords, **kwargs):
        return coords, kwargs
# MolecularZMatrixToRegularZMatrixConverter = MolecularZMatrixToRegularZMatrixConverter()
# MolecularZMatrixToRegularZMatrixConverter.register()



__all__ = [
    "SymbolicCaller",
    "AnalyticModelBase"
]

import numpy as np

class SympyShim:
    """
    Provides a loader that can either load sympy when requested
    or throw an error if it can't be loaded
    """
    sym = None
    @classmethod
    def _load_sympy(self):
        if self.sym is None:
            try:
                import sympy as sym
            except ImportError:
                raise ImportError("sympy must be installed to use analytic models")
            self.sym = sym
        return self.sym
    def __getattr__(self, item):
        return getattr(self._load_sympy(), item)
    # @classmethod
    # def _load_attr(self, item):
    #     return getattr(self._load_sympy(), item)
    # def tensorcontraction(self, ):
sym = SympyShim()

class SymbolicCaller:
    """
    Delegates to `__call__` through `__getitem__` for symbolics
    """
    def __init__(self, sym):
        self.sym = sym
    def __getitem__(self, item):
        if isinstance(item, int):
            return self.sym(item)
        else:
            return self.sym(*item)

class AnalyticModelBase:
    """
    Provides a base class for analytic models
    """
    sym = sym
    @classmethod
    def get_numeric_types(cls):
        return (int, float, np.integer, np.floating, np.ndarray, sym.Float, sym.Integer, sym.Rational)
    @classmethod
    def take_derivs(cls, expr, vars):
        """
        Takes derivatives of `expr` with respect to `vars` even if `expr` is an array

        :param expr:
        :type expr:
        :param vars:
        :type vars:
        :return:
        :rtype:
        """
        if isinstance(expr, list):
            expr = sym.Array(expr)
        return sym.Array([sym.diff(expr, v) for v in vars])

    @classmethod
    def eval_exprs(cls, expr, subs):
        """
        Evaluates `expr` with the given substitutions

        :param expr:
        :type expr:
        :param subs:
        :type subs:
        :return:
        :rtype:
        """
        # if isinstance(expr, sym.Array):
        #     return [cls.eval_exprs(e, subs) for e in expr]
        # elif isinstance(expr, (int, float, np.integer, np.floating)):
        #     return expr
        # else:
        if isinstance(expr, list):
            try:
                expr = sym.Array(expr)
            except ValueError:
                pass
        elif isinstance(expr, np.ndarray) and expr.dtype == object:
            expr = expr.tolist()
            try:
                expr = sym.Array(expr)
            except ValueError:
                pass
        if isinstance(expr, cls.get_numeric_types()):
            return expr
        elif not isinstance(expr, list):
            return expr.subs(subs)
        else:
            return [cls.eval_exprs(e, subs) for e in expr]

    @classmethod
    def symbol_list(cls, names, instance=None):
        """
        Gets a list of symbols for `names` with a given instance number

        :param names:
        :type names:
        :param instance:
        :type instance:
        :return:
        :rtype:
        """
        instance_tag = "" if instance is None else str(instance)
        return sym.symbols(" ".join(n+instance_tag for n in names))
    @staticmethod
    def symbolic_x(i):
        """
        Provides a symbolic representation of a position

        :param i:
        :type i:
        :return:
        :rtype:
        """
        return sym.Symbol('x[{i}]'.format(i=i), real=True)
    @staticmethod
    def symbolic_n(i, j, k):
        """
        Provides a symbolic representation of a normal to a plane

        :param i:
        :type i:
        :return:
        :rtype:
        """
        return sym.Symbol('n[{i},{j},{k}]'.format(i=i, j=j, k=k), real=True)
    @staticmethod
    def symbolic_m(i):
        """
        Provides a symbolic representation of a mass

        :param i:
        :type i:
        :return:
        :rtype:
        """
        return sym.Symbol('m[{i}]'.format(i=i), positive=True)
    @staticmethod
    def symbol(base, *args, **kwargs):
        if len(args) == 0:
            template = "{}"
        else:
            template = "{}["+",".join(["{}"]*len(args))+"]"
        return sym.Symbol(template.format(base, *args), **kwargs)
    @staticmethod
    def symbolic_r(i, j):
        """
        Provides a symbolic representation of a bond length

        :param i:
        :type i:
        :param j:
        :type j:
        :return:
        :rtype:
        """
        if j < i:
            i, j = j, i
        return sym.Symbol('r[{i},{j}]'.format(i=i, j=j), positive=True)
    @staticmethod
    def symbolic_a(i, j, k):
        """
        Provides a symbolic representation of a bond angle

        :param i:
        :type i:
        :param j:
        :type j:
        :param k:
        :type k:
        :return:
        :rtype:
        """
        if k < i:
            i, k = k, i
        return sym.Symbol('a[{i},{j},{k}]'.format(i=i, j=j, k=k), real=True)
    @staticmethod
    def symbolic_t(i, j, k, l):
        """
        Provides a symbolic representation of a dihedral angle

        :param i:
        :type i:
        :param j:
        :type j:
        :param k:
        :type k:
        :param l:
        :type l:
        :return:
        :rtype:
        """
        if l < i:
            i, j, k, l = l, k, j, i
        return sym.Symbol('t[{i},{j},{k},{l}]'.format(i=i, j=j, k=k, l=l), real=True)
    @staticmethod
    def symbolic_y(i, j, k, l):
        """
        Provides a symbolic representation of a book angle

        :param i:
        :type i:
        :param j:
        :type j:
        :param k:
        :type k:
        :param l:
        :type l:
        :return:
        :rtype:
        """
        if l < k:
            i, j, k, l = i, j, l, k
        return sym.Symbol('y[{i},{j},{k},{l}]'.format(i=i, j=j, k=k, l=l), real=True)

    @classmethod
    def infer_coord_type(cls, inds):
        if len(inds) == 0:
            return "d"
        elif len(inds) == 1:
            return "m"
        elif len(inds) == 2:
            return "r"
        elif len(inds) == 3:
            return "a"
        elif len(inds) == 4:
            return "t"
        elif len(inds) == 5 and inds[-1] == -1:
            return "y"
        else:
            raise ValueError("too many indices")
    @classmethod
    def var(cls, *args, coord_type=None):
        if coord_type is None:
            coord_type = cls.infer_coord_type(args)

        if coord_type == "d":
            return sym.Symbol("d", real=True)
        elif coord_type == "m":
            return cls.symbolic_m(*args)
        elif coord_type == "r":
            return cls.symbolic_r(*args)
        elif coord_type == "a":
            return cls.symbolic_a(*args)
        elif coord_type == "t":
            return cls.symbolic_t(*args)
        elif coord_type == "y":
            return cls.symbolic_y(*args)
        else:
            raise NotImplementedError("huh")

    @classmethod
    def reindex_symbol(cls, symbol, mapping, target_symbols=None):
        """
        Changes the indices on symbols using the given mapping

        :param symbol:
        :type symbol:
        :param mapping:
        :type mapping:
        :return:
        :rtype:
        """
        name = symbol.name
        for k,v in mapping.items():
            name = name.replace(str(k), str(k)+"$")
        qual_name = name
        for k,v in mapping.items():
            name = name.replace(str(k)+"$", str(v))
        if target_symbols is not None and name not in target_symbols:
            t, a = name.split("[")
            a = a.strip("]").split(",")
            inds = tuple(int(x) for x in a)
            if t == "r":
                test = cls.symbolic_r(inds[1], inds[0])
                if test.name in target_symbols:
                    name = test.name
            elif t == "a":
                test = cls.symbolic_a(inds[2], inds[1], inds[0])
                if test.name in target_symbols:
                    name = test.name
            elif t == "t":
                test = cls.symbolic_t(inds[3], inds[2], inds[1], inds[0])
                if test.name in target_symbols:
                    name = test.name

        return [sym.Symbol(qual_name, positive=symbol.is_positive, real=symbol.is_real), sym.Symbol(name, positive=symbol.is_positive, real=symbol.is_real)]

    @classmethod
    def lam(cls, i, j, k):
        """
        Provides the `lambda` expression from Frederick and Woywood

        :param i:
        :type i:
        :param j:
        :type j:
        :param k:
        :type k:
        :return:
        :rtype:
        """
        r12 = cls.symbolic_r(i, j)
        r23 = cls.symbolic_r(j, k)
        a = cls.symbolic_a(i, j, k)
        return 1/sym.sin(a)*(1/r12 - sym.cos(a)/r23)

    @classmethod
    def is_identity(cls, A):
        N = len(A)
        for i in range(N):
            m = len(A[i])
            if m != N:
                return False
            for j in range(m):
                if (j != i and A[i][j] != 0) or (j == i and A[i][j] != 1):
                    return False
        return True
    @classmethod
    def transpose(cls, A):
        if isinstance(A, list):
            A = sym.Array(A)
        return sym.transpose(A)
    @classmethod
    def dot(cls, a, b, axes=None):
        if isinstance(a, (list, sym.Matrix)):
            a = sym.Array(a)
        if isinstance(b, (list, sym.Matrix)):
            b = sym.Array(b)
        if isinstance(a, np.ndarray) and isinstance(b, np.ndarray):
            if axes is None:
                return np.dot(a, b)
            else:
                return np.tensordot(a, b, axes=axes)
        else:
            if isinstance(a, np.ndarray):
                a = sym.Array(a)
            if isinstance(b, np.ndarray):
                b = sym.Array(b)
            if axes is None:
                axes = (a.rank()-1, a.rank())
            else:
                axa, axb = axes
                if isinstance(axa, int) and axa < 0:
                    axa = a.rank() + axa + 1
                if isinstance(axb, int):
                    if axb < 0:
                        axb = b.rank() + axb + 1
                    axb += a.rank()-1
                else:
                    ar = a.rank()-1
                    s = b.rank() + 1
                    axb = [ar + (x + s if x < 0 else x) for x in axb]
                axes = [axa, axb]
            return sym.tensorcontraction(sym.tensorproduct(a, b), axes)
    @classmethod
    def contract(cls, a, axes):
        if isinstance(a, list):
            a = sym.Array(a)
        return sym.tensorcontraction(a, axes)

    @classmethod
    def transform_coordinates(cls, rotation, coord_vec=None, coord_name_fmt="q{id}[{num}]"):
        if coord_vec is None:
            import random

            rand_id = random.randint(100, 999)
            coord_vec = [
                coord_name_fmt.format(
                    id=rand_id,
                    num=i+1
                ) for i in range(len(rotation[0]))
            ]
        coord_vec = [sym.Symbol(s) if isinstance(s, str) else s for s in coord_vec]

        return coord_vec, cls.dot(rotation, coord_vec)



import abc, numpy as np, functools
from McUtils.Data import AtomData, UnitsData
import McUtils.Numputils as nput

from ..Molecools import StructuralProperties

from .Evaluators import *
from .Interpolation import *


__all__ = [
    "DGBCoords",
    "DGBCartesians",
    "DGBInternals",
    "DGBWatsonModes",
]

class DGBCoords(metaclass=abc.ABCMeta):

    @property
    @abc.abstractmethod
    def centers(self) -> 'np.ndarray':
        ...
    @property
    def shape(self):
        return self.centers.shape

    @property
    @abc.abstractmethod
    def kinetic_energy_evaluator(self) -> 'DGBKineticEnergyEvaluator':
        ...

    @property
    @abc.abstractmethod
    def pairwise_potential_evaluator_type(self) -> 'type[DGBPairwisePotentialEvaluator]':
        ...
    def pairwise_potential_evaluator(self, potential_functions) -> 'DGBPairwisePotentialEvaluator':
        if not isinstance(potential_functions, dict):
            raise ValueError("don't know what to do with pairwise functions {}".format(potential_functions))
        if 'functions' in potential_functions:
            opts = potential_functions.copy()
            potential_functions = potential_functions['functions']
            del opts['functions']
        else:
            opts = {}
        return self.pairwise_potential_evaluator_type(self, potential_functions, **opts)
    @abc.abstractmethod
    def __getitem__(self, item) -> 'DGBCoords':
        ...
    def take_indices(self, subinds) -> 'DGBCoords':
        return self[:, subinds] # default
    def drop_indices(self, subinds) -> 'DGBCoords':
        remaining = np.setdiff1d(np.arange(self.centers.shape[-1]), subinds)
        return self.take_indices(remaining) # default
    @abc.abstractmethod
    def gmatrix(self, coords:np.ndarray) -> np.ndarray:
        ...

    @classmethod
    def embedded_mode_function(cls, func, modes, masses=None):
        if isinstance(func, DGBInterpolator):
            centers = func.centers
            pot_vals = func.derivs
            opts = func.opts
            centers = DGBWatsonModes.embed_coords(centers, modes)
            pot_vals = DGBWatsonModes.embed_derivs(pot_vals, modes)
            return DGBWatsonInterpolator(centers, pot_vals, modes, **opts)
        else:
            @functools.wraps(func)
            def embedded_function(mode_coords, deriv_order=None):
                carts = DGBWatsonModes.unembed_coords(mode_coords, modes, masses)
                vals = func(carts, deriv_order=deriv_order)
                if deriv_order is not None:
                    vals = DGBWatsonModes.embed_derivs(vals, modes)
                return vals
            return embedded_function

    @classmethod
    def embedded_subcoordinate_function(cls, func, sel, ndim):
        if isinstance(func, DGBInterpolator):
            raise NotImplementedError("reembedding of interpolator not supported")
        @functools.wraps(func)
        def embedded_function(subcoords, deriv_order=None):
            full_shape = (subcoords.shape[0], ndim)
            if sel is None:
                coords = subcoords.reshape(full_shape)
            else:
                coords = np.zeros(full_shape)
                coords[:, sel] = subcoords
            vals = func(coords, deriv_order=deriv_order)
            if sel is not None and deriv_order is not None:
                fshape = vals[0].ndim
                _ = []
                for n, d in enumerate(vals):
                    for j in range(n):
                        d = np.take(d, sel, axis=j + fshape)
                    _.append(d)
                vals = _
            return vals

        return embedded_function

    @classmethod
    def embedded_cartesian_function(cls, func, atom_sel, xyz_sel, natoms, ndim):
        if (atom_sel is not None or xyz_sel is not None):
            full_sel = np.arange(natoms * ndim).reshape(natoms, ndim)
            if atom_sel is None:
                full_sel = full_sel[:, xyz_sel]
            elif xyz_sel is None:
                full_sel = full_sel[atom_sel, :]
            else:
                full_sel = full_sel[np.ix_(atom_sel, xyz_sel)]
            flat_sel = full_sel.flatten()
        else:
            flat_sel = None

        if flat_sel is None:
            return func

        if isinstance(func, DGBInterpolator):
            centers = func.centers
            centers = centers.reshape((centers.shape[0], natoms, ndim))
            pot_vals = func.derivs
            opts = func.opts

            if atom_sel is None:
                centers = centers[..., xyz_sel]
            elif xyz_sel is None:
                centers = centers[..., atom_sel, :]
            else:
                centers = centers[..., np.ix_(atom_sel, xyz_sel)]
            fshape = pot_vals[0].ndim
            _ = []
            for n, d in enumerate(pot_vals):
                for j in range(n):
                    d = np.take(d, flat_sel, axis=j + fshape)
                _.append(d)
            pot_vals = _
            if isinstance(func, DGBWatsonInterpolator):
                return DGBCartesianWatsonInterpolator(centers, pot_vals, func.modes, **opts)
            else:
                return DGBInterpolator(centers, pot_vals, **opts)
        else:
            if flat_sel is None:
                return func
            @functools.wraps(func)
            def embedded_function(subcart_coords, deriv_order=None, flat_sel=flat_sel):
                full_shape = (subcart_coords.shape[0], natoms, ndim)
                if subcart_coords.shape == full_shape:
                    carts = subcart_coords
                else:
                    subcart_coords = np.reshape(
                        subcart_coords,
                        (
                            subcart_coords.shape[0],
                            len(atom_sel) if atom_sel is not None else natoms,
                            len(xyz_sel) if xyz_sel is not None else ndim
                        )
                    )
                    if atom_sel is None and xyz_sel is None:
                        carts = subcart_coords.reshape(full_shape)
                    else:
                        carts = np.zeros(full_shape)
                        if atom_sel is None:
                            carts[..., xyz_sel] = subcart_coords
                        elif xyz_sel is None:
                            carts[..., atom_sel, :] = subcart_coords
                        else:
                            carts[..., np.ix_(atom_sel, xyz_sel)] = subcart_coords
                vals = func(carts, deriv_order=deriv_order)
                if deriv_order is not None:
                    fshape = vals[0].ndim
                    _ = []
                    for n, d in enumerate(vals):
                        for j in range(n):
                            d = np.take(d, flat_sel, axis=j + fshape)
                        _.append(d)
                    vals = _
                return vals

        return embedded_function

    class DGBEmbeddedFunction:
        def __init__(self, embedded_function, original_function, coords):
            self.og_fn = original_function
            self.base_coords = coords
            self.embed_fn = embedded_function
        def __call__(self, coords, deriv_order=None):
            return self.embed_fn(coords, deriv_order=deriv_order)
    @abc.abstractmethod
    def embed_function(self, fn) -> 'DGBEmbeddedFunction':
        ...

    def as_cartesians(self) -> 'tuple[DGBCartesians, tuple[np.ndarray, np.ndarray]]':
        raise NotImplementedError("{} can't be converted to Cartesians".format(
            type(self).__name__
        ))

class DGBCartesians(DGBCoords):
    def __init__(self, coords, masses, *, natoms=None, atom_sel=None, ndim=None, xyz_sel=None):
        coords = np.asanyarray(coords)
        masses = np.asanyarray(masses)
        if masses.ndim > 1:
            raise ValueError("expected a vector of masses...")
        coords = coords.reshape((len(coords), len(masses), -1))

        self.coords = coords
        self.masses = masses
        self.natoms = len(masses) if natoms is None else natoms# the original numebr of atoms
        self.ndim = self.coords.shape[-1] if ndim is None else ndim
        self.atom_sel = atom_sel
        self.xyz_sel = xyz_sel
    @property
    def centers(self):
        return self.coords.reshape((self.coords.shape[0], self.coords.shape[1]*self.coords.shape[2]))
    @property
    def cart_shape(self):
        return self.coords.shape
    @property
    def kinetic_energy_evaluator(self):
        return DGBCartesianEvaluator(
            np.broadcast_to(self.masses[:, np.newaxis], self.coords.shape[1:]).flatten()
        )
    @property
    def pairwise_potential_evaluator_type(self):
        return DGBCartesianPairwiseEvaluator
    @classmethod
    def resolve_masses(cls, coords, masses=None, atoms=None):
        if masses is None:
            if atoms is not None:
                atoms = [
                    AtomData[a, "Mass"] * UnitsData.convert("AtomicMassUnits", "AtomicUnitOfMass")
                        if isinstance(a, str) else
                    a
                    for a in atoms
                ]
                masses = np.array(atoms)
            else:
                raise ValueError("either atoms or masses must not be `None`")

        if isinstance(masses, (int, np.integer, float, np.floating)):
            masses = [masses] * coords.shape[1]
        return masses
    @classmethod
    def from_cartesians(cls, centers, masses=None, atoms=None):
        centers = np.asanyarray(centers)
        masses = cls.resolve_masses(centers, masses=masses, atoms=atoms)
        return cls(centers, masses)

    def infer_shape_sel(self, selector):
        n, na, nd = self.coords.shape
        test = np.arange(na*nd).reshape((1, na, nd))
        inferrable = np.broadcast_to(test, self.coords.shape)[selector][0]

        test_atoms = test[0, :, 0] // nd
        sub_atoms = inferrable[:, 0] // nd
        # we map the atoms back to their positions in the original list
        atom_sel = np.searchsorted(test_atoms, sub_atoms)
        if len(atom_sel) == na and np.all(atom_sel == np.arange(na)):
            atom_sel = None

        test_xyz = test[0, 0] % nd
        sub_xyz = inferrable[0] % nd
        xyz_sel, _ = nput.find(test_xyz, sub_xyz)
        if len(xyz_sel) == nd and np.all(xyz_sel == np.arange(nd)):
            xyz_sel = None

        return atom_sel, xyz_sel
    @staticmethod
    def _merge_sel(new_sel, old_sel):
        if old_sel is not None:
            new_sel = np.asanyarray(old_sel)[new_sel,]
        return new_sel
    def __getitem__(self, item) -> 'DGBCartesians':
        c = self.coords[item]
        if not isinstance(c, np.ndarray) or c.ndim != 3:
            raise ValueError("bad slice {}".format(item))

        if c.ndim < 3:
            raise ValueError("zero-length slice?")

        if c.shape[1:] == self.coords.shape[1:]:
            return type(self)(
                c,
                self.masses,
                atom_sel=self.atom_sel,
                xyz_sel=self.xyz_sel,
                natoms=self.natoms,
                ndim=self.ndim
            )
        else:
            atom_sel, xyz_sel = self.infer_shape_sel(item)
            return type(self)(
                c,
                self.masses if atom_sel is None else self.masses[atom_sel,],
                atom_sel=self._merge_sel(atom_sel, self.atom_sel),
                xyz_sel=self._merge_sel(xyz_sel, self.xyz_sel),
                natoms=self.natoms,
                ndim=self.ndim
            )
    def take_indices(self, subinds):
        subinds = np.asanyarray(subinds)
        atom_inds = np.unique(subinds // self.coords.shape[2])
        xyz_inds = np.unique(subinds % self.coords.shape[2])
        return self[:, atom_inds, :][:, :, xyz_inds]
    def embed_function(self, function):
        """
        Embeds assuming we got a function in Cartesians _before_ any selections happened

        :param function:
        :return:
        """
        return self.DGBEmbeddedFunction(
            self.embedded_cartesian_function(
                function,
                self.atom_sel,
                self.xyz_sel,
                self.natoms,
                self.ndim
            ),
            function,
            self
        )

    def gmatrix(self, coords:np.ndarray) -> np.ndarray:
        mass_spec = np.broadcast_to(self.masses[:, np.newaxis], (len(self.masses), self.coords.shape[2])).flatten()
        mass_spec = np.diag(1 / mass_spec)
        return np.broadcast_to(mass_spec[np.newaxis], (coords.shape[0],) + mass_spec.shape)
class DGBInternals(DGBCoords):
    def __init__(self, coords, gmat_function=None, vprime_function=None):
        raise NotImplementedError('internals coming soon?')
class DGBWatsonModes(DGBCoords):
    def __init__(self, coords, modes,
                 *,
                 coriolis_inertia_function=None,
                 masses=None,
                 subselection=None
                 ):
        self.coords = coords
        self.masses = masses
        self.modes = modes
        self.subsel = subselection
        self.ci_func = coriolis_inertia_function

    @property
    def centers(self):
        return self.coords
    @property
    def kinetic_energy_evaluator(self):
        return DGBWatsonEvaluator(self.modes, self.ci_func)
    @property
    def pairwise_potential_evaluator_type(self):
        return DGBWatsonPairwiseEvaluator
    @staticmethod
    def zeta_momi(watson_coords, modes, masses):

        carts = DGBWatsonModes.unembed_coords(watson_coords, modes, masses=masses)
        if carts.shape[-1] < 3:
            # pad with zeros
            carts = np.concatenate(
                [
                    carts,
                    np.zeros(carts.shape[:2] + (3 - carts.shape[-1],))
                ],
                axis=-1
            )

        # scaled_modes = modes.matrix.T / np.sqrt(modes.freqs[:, np.newaxis]) # used in the VPT version?
        zeta, (B_e, eigs) = StructuralProperties.get_prop_coriolis_constants(carts,
                                                                             modes.matrix.T,
                                                                             masses
                                                                             )

        return zeta, B_e

    @classmethod
    def default_coriolis_inertia_function(cls, modes, masses):
        def coriolis_inertia_function(watson_coords, *, modes=modes, masses=masses, deriv_order=None):
            if deriv_order is not None:
                raise NotImplementedError("don't have support for Coriolis derivatives in DGB")
            zeta, mom_i = cls.zeta_momi(watson_coords, modes, masses)
            # freqs = modes.freqs
            # freq_term = np.sqrt(freqs[np.newaxis, :] / freqs[:, np.newaxis])
            # zeta = zeta * freq_term[np.newaxis, np.newaxis]
            bad_moms = np.where(mom_i <= 1)
            mom_i[bad_moms] = 1
            B_e = 1 / (2 * mom_i)  # * UnitsData.convert("AtomicMassUnits", "AtomicUnitOfMass"))
            B_e[bad_moms] = 0
            return np.sum(B_e, axis=-1), sum(
                B_e[:, a, np.newaxis, np.newaxis, np.newaxis, np.newaxis]
                 * zeta[:, a, :, :, np. newaxis, np.newaxis]*zeta[:, a, np. newaxis, np.newaxis, :, :]
                for a in range(B_e.shape[1])
            )

        return coriolis_inertia_function

    @classmethod
    def embed_coords(cls, carts, modes, shift=True):
        if shift:
            flat_carts = (carts - modes.origin[np.newaxis]).reshape((len(carts), -1))
        else:
            flat_carts = (carts).reshape((len(carts), -1))
        return (flat_carts[:, np.newaxis, :] @ modes.inverse.T[np.newaxis]).reshape(
            flat_carts.shape[0],
            modes.matrix.shape[1]
        )
    @classmethod
    def unembed_coords(cls, mode_coords, modes, masses=None, shift=True):
        origin = modes.origin
        carts = (mode_coords[:, np.newaxis, :] @ modes.matrix.T[np.newaxis]).reshape(
            mode_coords.shape[:1] + origin.shape
        )
        if shift:
            carts = carts + origin[np.newaxis, :, :]
        # if masses is not None:
        #     carts = carts.reshape((carts.shape[0], len(masses), -1))
        #     carts = StructuralProperties.get_eckart_embedded_coords(
        #         masses,
        #         origin,
        #         carts
        #     )
        return carts
    @classmethod
    def embed_derivs(cls, derivs, modes):
        fshape = derivs[0].ndim
        _ = []
        for n, d in enumerate(derivs):
            for j in range(n):
                d = np.tensordot(d, modes.matrix, axes=[fshape, 0])
            _.append(d)
        return _
    @classmethod
    def from_cartesians(cls,
                        coords,
                        modes,
                        masses=None,
                        coriolis_inertia_function=None
                        ):
        coords = cls.embed_coords(coords, modes)
        if coriolis_inertia_function is None:
            coriolis_inertia_function = cls.default_coriolis_inertia_function(modes, masses)
        return cls(coords,
                   modes,
                   coriolis_inertia_function=coriolis_inertia_function,
                   masses=masses
                   )
    def as_cartesians(self, masses=None) -> 'tuple[DGBCartesians, tuple[np.ndarray, np.ndarray]]':
        mode_coords = self.coords
        modes = self.modes
        if masses is None:
            masses = self.masses

        carts = self.unembed_coords(mode_coords, modes, self.masses)
        if masses is None:
            raise ValueError("need `masses` to be able to convert back to Cartesians")
        carts = carts.reshape((carts.shape[0], len(masses), -1))

        return DGBCartesians(carts, masses), (modes.matrix, modes.inverse)

    def __getitem__(self, item):
        c = self.coords[item]
        if not isinstance(c, np.ndarray) or c.ndim != self.coords.ndim:
            raise ValueError("bad slice {}".format(item))
        if c.shape[1] != self.coords.shape[1]:
            # means we took some subselection of modes
            test_sel = np.broadcast_to(
                np.arange((self.coords.shape[1]))[np.newaxis],
                self.coords.shape
            ) # something that can be indexed the same way
            subsel = test_sel[item][0]
            modes = self.modes[subsel]
            ci_func = self.embedded_subcoordinate_function(
                self.ci_func,
                subsel,
                self.coords.shape[1]
            )
        else:
            modes = self.modes
            subsel = None
            ci_func = self.ci_func
        if self.subsel is not None:
            subsel = self.subsel[subsel]
        return type(self)(
            c,
            modes,
            coriolis_inertia_function=ci_func,
            masses=self.masses,
            subselection=subsel
        )
    def gmatrix(self, coords:np.ndarray) -> np.ndarray:
        base_spec = np.eye(coords.shape[1])
        return np.broadcast_to(base_spec[np.newaxis], (coords.shape[0],) + base_spec.shape)

    def embed_function(self, fn):
        return self.DGBEmbeddedFunction(
            self.embedded_mode_function(  # a function that takes in normal mode coordinates
                fn,
                self.modes,
                masses=self.masses
            ),
            fn,
            self
        )



import abc

__all__ = [
]
class VibronicState:
    """
    Abstract base class for representing and electronic-vibrational product state
    """
    @abc.abstractmethod
    def get_overlaps(self):
        ...


class VibronicModel:
    """
    Provides an abstraction for calculating vibronic spectra
    """

    def __init__(self, *vibronic_states:'VibronicState', **opts):
        self.states = vibronic_states
        self.opts = opts

    def calculate_spectrum(self):
        ...